"""Test metadata updates."""

import doranet as dn


def test_meta_update():
    engine = dn.create_engine()
    network = engine.new_network()
    network.add_mol(engine.mol.rdkit("C#C"), meta={"gen": 0})
    network.add_op(engine.op.rdkit("[C:1]#[C:2]>>[*:1]=[*:2]"))
    network.add_op(engine.op.rdkit("[C:1]#[C:2]>>[*:1]-[*:2]"))
    network.add_op(engine.op.rdkit("[C:1]=[C:2]>>[*:1]-[*:2]"))
    network.add_op(engine.op.rdkit("[C:1]-[C:2]>>[*:1].[*:2]"))
    network.add_op(engine.op.rdkit("[C:1]-[C:2]>>[*:1]=[*:2]"))

    strat = engine.strat.cartesian(network)

    strat.expand(reaction_plan=dn.metacalc.GenerationCalculator("gen"))

    assert network.mols.meta(dn.interfaces.MolIndex(0), ("gen",))["gen"] == 0
    assert network.mols.meta(dn.interfaces.MolIndex(1), ("gen",))["gen"] == 1
    assert network.mols.meta(dn.interfaces.MolIndex(2), ("gen",))["gen"] == 1
    assert network.mols.meta(dn.interfaces.MolIndex(3), ("gen",))["gen"] == 2  # noqa: PLR2004



"""Contains classes which define and implement utility functions."""

import collections.abc
import dataclasses
import itertools
import typing

from doranet import interfaces


@dataclasses.dataclass(frozen=True, slots=True)
class RxnTrackerDepthFirstNetwork(interfaces.RxnTrackerNetwork):
    """Implements RxnTrackerNetwork interface."""

    network: interfaces.ChemNetwork

    def _getchains(
        self,
        cur_gen_mols: collections.abc.Collection[interfaces.MolIndex],
        prev_gens_mols: typing.Optional[set[interfaces.MolIndex]] = None,
        prev_gens_rxns: typing.Optional[set[interfaces.RxnIndex]] = None,
        reagent_table: typing.Optional[
            collections.abc.Container[interfaces.MolIndex]
        ] = None,
        fail_on_unknown_reagent: bool = False,
        depth: typing.Optional[int] = None,
    ) -> collections.abc.Generator[
        list[frozenset[interfaces.RxnIndex]], None, None
    ]:
        network = self.network
        n_mols = len(network.mols)
        # n_rxns = len(network.rxns)
        if depth is not None:
            if depth <= 0:
                return
            new_depth = depth - 1
        else:
            new_depth = None
        if len(cur_gen_mols) == 0:
            yield []
            return
        if prev_gens_mols is None:
            prev_gens_mols = set()
        if prev_gens_rxns is None:
            prev_gens_rxns = set()
        if reagent_table is None:
            reagent_table = set()
        rxnsets: list[list[interfaces.RxnIndex]] = []
        for mol in cur_gen_mols:
            if mol in reagent_table:
                continue
            elif mol >= n_mols:
                rxnsets.append([])
                continue
            newrxnset: list[interfaces.RxnIndex] = [
                rxn
                for rxn in network.producers(mol)
                if rxn not in prev_gens_rxns
                and all(
                    mol not in prev_gens_mols
                    for mol in network.rxns[rxn].reactants
                )
            ]
            rxnsets.append(newrxnset)
        if not fail_on_unknown_reagent:
            rxnsets = [rxnset for rxnset in rxnsets if len(rxnset) > 0]
            if len(rxnsets) == 0:  # these two lines
                return  # close the loophole
        if len(rxnsets) == 0:
            yield []
            return
        tested_combos: set[frozenset[interfaces.RxnIndex]] = set()
        for rxntuple in itertools.product(*rxnsets):
            rxncombo: frozenset[interfaces.RxnIndex] = frozenset(rxntuple)
            if rxncombo in tested_combos:
                continue
            else:
                tested_combos.add(frozenset(rxncombo))
            required_reagents = set(
                mol
                for mol in itertools.chain(
                    *(network.rxns[rxn].reactants for rxn in rxncombo)
                )
                if mol not in reagent_table
            )
            if len(required_reagents) == 0:
                yield [rxncombo]
                continue
            for path in self._getchains(
                required_reagents,
                prev_gens_mols.union(cur_gen_mols),
                prev_gens_rxns.union(rxncombo),
                reagent_table,
                fail_on_unknown_reagent,
                new_depth,
            ):
                path.append(rxncombo)
                yield path

    def getParentChains(
        self,
        target: interfaces.MolIndex,
        reagent_table: collections.abc.Container[interfaces.MolIndex] = tuple(),
        fail_on_unknown_reagent: bool = False,
        max_depth: typing.Optional[int] = None,
    ) -> collections.abc.Generator[
        list[frozenset[interfaces.RxnIndex]], None, None
    ]:
        if fail_on_unknown_reagent and not reagent_table:
            raise ValueError(
                "reagent table must be specified if fail_on_unknown_reagent is "
                "True"
            )
        return (
            path
            for path in self._getchains(
                [target],
                reagent_table=reagent_table,
                fail_on_unknown_reagent=fail_on_unknown_reagent,
                depth=max_depth,
            )
        )


T = typing.TypeVar("T")


def logreduce(
    function: collections.abc.Callable[[T, T], T],
    iterable: collections.abc.Iterable[T],
) -> T:
    """
    logreduce(function, iterable[, initial]) -> value.

    Apply a function of two arguments (satisfying the associative property)
    cumulatively to the items of a sequence or iterable, from left to right, so
    as to reduce the iterable to a single value.  For example, reduce(lambda x,
    y: x+y, [1, 2, 3, 4, 5]) calculates (((1+2)+(3+4))+5).  If initial is
    present, it is placed before the items of the iterable in the calculation,
    and serves as a default when the iterable is empty.

    Memory: maximum of log2(n) objects of iterable stored vs 2 for reduce()
    Speed: log2(n) calls to function() vs n for reduce()
    """
    i = iter(iterable)
    try:
        r_val, stop = _logreduce(function, i, 0)
    except StopIteration:
        raise TypeError(
            "logreduce() of empty iterable with no initial value"
        ) from None
    n = 0
    try:
        while not stop:
            n += 1
            new_val, stop = _logreduce(function, i, n - 1)
            r_val = function(r_val, new_val)
    except StopIteration:
        ...
    return r_val


def _logreduce(
    f: collections.abc.Callable[[T, T], T],
    i: collections.abc.Iterator[T],
    n: int,
) -> tuple[T, bool]:
    if n == 0:
        x = next(i)
        try:
            y = next(i)
        except StopIteration:
            return x, True
        return f(x, y), False
    x, stop = _logreduce(f, i, n - 1)
    if stop:
        return x, True
    try:
        y, stop = _logreduce(f, i, n - 1)
    except StopIteration:
        return x, True
    return f(x, y), False


# def getFigures(
#     target_smiles: str,
#     mol_smiles: tuple[str, ...],
#     num_gens: int,
#     engine: interfaces.NetworkEngine,
#     rxn_lib: interfaces.ObjectLibrary[interfaces.RxnDatBase],
#     job_name: str,
# ):
#     """
#     Create and display figures of reaction pathways.

#     If the target is found at the end of the expansion, create figures of the
#     reaction routes. Each figure contains all generations of one route. Also
#     prints reaction operators, reactants, and products. Figures are displayed,
#     also saved in the working folder. Based on the example code for using
#     RxnTrackerDepthFirst(rxn_lib).  Currently need to have the font file
#     'OpenSans-Regular.ttf' in the working folder.  May not work due to legacy
#     code source.

#     Parameters
#     ----------
#     target_smiles : str
#         Target molecule SMILES.
#     mol_smiles : tuple[str,...]
#         SMILES of starting chemicals.
#     num_gens : int
#         Number of expansion generations performed.
#     engine : NetworkEngine
#         Network Engine used to create data objects.
#     rxn_lib : ObjectLibrary[RxnDatBase]
#         Reaction library where reactions are contained.
#     job_name: str
#         Name of the job.
#     """
#     doubletracker = RxnTrackerDepthFirst(rxn_lib)
#     route = 1
#     for chain in doubletracker.getParentChains(
#         engine.Mol(target_smiles).uid,
#         reagent_table=[engine.Mol(smiles).uid for smiles in mol_smiles],
#         max_depth=num_gens,
#     ):
#         g = 0
#         rxn_image_list = []
#         for gen in chain:
#             print(f"Generation {g}:")
#             for rxnid in gen:
#                 print(rxn_lib[rxnid])
#                 rxn_smiles_str = ""
#                 for molecule in rxn_lib[rxnid].reactants:
#                     if not isinstance(molecule, str):
#                         raise TypeError(
#                             f"""Identifier {molecule} of type {type(molecule)}
#                                 must be SMILES string."""
#                         )
#                     rxn_smiles_str += molecule
#                     rxn_smiles_str += "."
#                 rxn_smiles_str += ">>"
#                 for molecule in rxn_lib[rxnid].products:
#                     if not isinstance(molecule, str):
#                         raise TypeError(
#                             f"""Identifier {molecule} of type {type(molecule)}
#                                 must be SMILES string."""
#                         )
#                     rxn_smiles_str += molecule
#                     rxn_smiles_str += "."
#                 rxn1 = rdkit.Chem.rdchemreactions.ReactionFromSmarts(
#                     rxn_smiles_str, useSmiles=True
#                 )
#                 img1 = rdkit.Chem.Draw.ReactionToImage(rxn1)
#                 msg = f"Route {route}, Generation {g}"
#                 img_w, img_h = img1.size
#                 I1 = PIL.ImageDraw.Draw(img1)
#                 myFont = PIL.ImageFont.truetype("OpenSans-Regular.ttf", 25)
#                 new_box = I1.textbbox((0, 0), msg, font=myFont)
#                 I1.text(
#                     ((img_w - new_box[2]) / 2, 0),
#                     msg,
#                     font=myFont,
#                     fill="black",
#                 )
#                 rxn_image_list.append(img1)
#             g += 1
#         max_width = 0
#         total_hight = 0
#         for img in rxn_image_list:
#             total_hight += img.size[1]
#             max_width = max(max_width, img.size[0])
#         combined_image = PIL.Image.new("RGB",
#                                        (max_width, total_hight),
#                                        "white")
#         current_hight = 0
#         for img in rxn_image_list:
#             combined_image.paste(
#                 img, ((max_width - img.size[0]) // 2, current_hight)
#             )
#             current_hight += img.size[1]
#         IPython.display(combined_image)  # type: ignore
#         combined_image.save(f"{job_name} route {route}.png")
#         route += 1



"""DORAnet module for post-processing."""

__all__ = [
    "pathway_finder",
    "pathway_ranking",
    "pathway_visualization",
    "pretreat_networks",
    "one_step",
]

from .post_processing import (
    one_step,
    pathway_finder,
    pathway_ranking,
    pathway_visualization,
    pretreat_networks,
)



from datetime import datetime

from IPython.display import display

todays_date = str(datetime.now())

import dataclasses
import re
import sys
import typing
from typing import Optional, Tuple, Union

from rdkit import Chem
from rdkit.Chem import AllChem, Descriptors, Draw

import doranet
from doranet import interfaces, metadata

print(f"Last update on {todays_date}")
print(sys.version)
print("rdkit version: ", Chem.rdBase.rdkitVersion)

## Initialize doranet engine
engine = doranet.create_engine()
network = engine.new_network()


def print_time(time_end, time_start):
    if time_end - time_start > 120:  # more than 120 seconds (~ 2 min)
        print(f"Executed in {(time_end - time_start)/60:.3f} min")
    else:
        print(f"Executed in {(time_end - time_start):.3f} sec")


def update_isotope_smiles(
    input_string: str = None,
    special_species_list_not_to_update_isotope: list = None,
):
    if special_species_list_not_to_update_isotope is None:
        special_species_list_not_to_update_isotope = ["O=O"]

    if input_string in special_species_list_not_to_update_isotope:
        output_string = input_string
    else:
        input_string_addHs = Chem.MolToSmiles(
            Chem.AddHs(Chem.MolFromSmiles(input_string))
        )

        def Hydrogen_Count(match):
            nonlocal hydrogen_counter
            hydrogen_counter += 1
            return f"[{hydrogen_counter}H]"

        hydrogen_counter = 0
        output_string = re.sub(r"\[H\]", Hydrogen_Count, input_string_addHs)

    return output_string


def canonical_smiles_parsing(
    input_string: str = None, visualize_mol: bool = None
):
    output_string = re.sub(r"\[(\d+)H\]", r"[H]", input_string)
    output_string = Chem.RemoveHs(Chem.MolFromSmiles(output_string))

    def show_atom_number(mol, label="atomNote", figsize=(300, 200)):
        for atom in mol.GetAtoms():
            atom.SetProp(label, str(atom.GetIdx()))

        AllChem.Compute2DCoords(mol)
        display(Draw.MolToImage(mol, size=figsize))

    if visualize_mol:
        show_atom_number(output_string)

    return Chem.MolToSmiles(output_string)


def write_to_file(
    output_to_save: list | dict | tuple = None, file_path: str = None
):
    if isinstance(output_to_save, list | tuple):
        with open(file_path, "w") as txt_file:
            for idx, value in enumerate(output_to_save):
                txt_file.write(str(value) + "\n")

    elif isinstance(output_to_save, dict):
        with open(file_path, "w") as txt_file:
            for idx, value in output_to_save.items():
                if len(output_to_save) < 1e5:
                    txt_file.write(f"{idx:<7}" + str(value) + "\n")
                elif len(output_to_save) < 1e8:
                    txt_file.write(f"{idx:<10}" + str(value) + "\n")
                elif len(output_to_save) < 1e11:
                    txt_file.write(f"{idx:<13}" + str(value) + "\n")
                elif len(output_to_save) < 1e14:
                    txt_file.write(f"{idx:<17}" + str(value) + "\n")
                elif len(output_to_save) > 1e14:
                    txt_file.write(f"{idx:<25}" + str(value) + "\n")


def write_rxn(reaction_data_list: list = None):
    _rxn_info, rxn_output = [], []

    for i in reaction_data_list:
        reactants_list = [str(i) for i in i[0]]
        prod_list = [str(i) for i in i[1]]
        rxn_abbrev = i[2]

        if len(reactants_list) > 1:
            reactant_text = ", ".join(reactants_list)
        else:
            reactant_text = reactants_list[0]

        if len(prod_list) > 1:
            product_text = ", ".join(prod_list)
        else:
            product_text = prod_list[0]

        rxn_output.append(f"{reactant_text} : {product_text} : {rxn_abbrev} ;")

    return rxn_output


@typing.final
@dataclasses.dataclass(frozen=True)
class RadicalPosition(metadata.ReactionFilterBase):
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        for idx, mol in enumerate(recipe.reactants):
            if (
                Chem.Descriptors.NumRadicalElectrons(mol.item.rdkitmol)
                != recipe.operator.meta["radical_tuple"][idx]
            ):
                return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(operator_keys={"radical_tuple"})


@typing.final
@dataclasses.dataclass(frozen=True)
class OxygenAppearance(metadata.ReactionFilterBase):
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        if recipe.operator.meta["oxygen_presence_tuple"] is not None:
            for idx, mol in enumerate(recipe.reactants):
                if (
                    recipe.operator.meta["oxygen_presence_tuple"][idx]
                    is not None
                ) and (
                    mol.meta["oxygen_presence"]
                    != recipe.operator.meta["oxygen_presence_tuple"][idx]
                ):
                    return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(operator_keys={"oxygen_presence_tuple"})


@typing.final
@dataclasses.dataclass(frozen=True)
class ComparingCarbonChainLength(metadata.ReactionFilterBase):
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        if (
            recipe.operator.meta["C_atom_check"] is not None
            and (
                len(set([mol.meta["carbon_num"] for mol in recipe.reactants]))
                == 1
            )
            and (recipe.operator.meta["C_atom_check"] is True)
        ):
            return False

        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(operator_keys={"C_atom_check"})


@typing.final
@dataclasses.dataclass(frozen=True)
class CalculateCarbonChainLength(metadata.ReactionFilterBase):
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        if recipe.operator.meta["max_permit_carbon"] is not None:
            for idx, mol in enumerate(recipe.reactants):
                if (
                    recipe.operator.meta["max_permit_carbon"][idx] is not None
                ) and (
                    mol.meta["carbon_num"]
                    >= recipe.operator.meta["max_permit_carbon"][idx] + 1
                ):
                    return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(operator_keys={"max_permit_carbon"})


@typing.final
@dataclasses.dataclass(frozen=True)
class CheckOxygenNumberReactants(metadata.ReactionFilterBase):
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        if recipe.operator.meta["OXY_ABSTRACT_COUNT"] is not None:
            for idx, mol in enumerate(recipe.reactants):
                if recipe.operator.meta["OXY_ABSTRACT_COUNT"][idx] is not None:
                    if (
                        mol.meta["oxygen_num"]
                        >= recipe.operator.meta["OXY_ABSTRACT_COUNT"][idx] + 1
                    ):
                        return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(operator_keys={"OXY_ABSTRACT_COUNT"})


@typing.final
@dataclasses.dataclass(frozen=True)
class CheckSpeciesSimilarity(metadata.ReactionFilterBase):
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        if recipe.operator.meta["species_similarity_check"] is not None:
            reactant_list = sorted(
                list(
                    set(
                        [
                            canonical_smiles_parsing(
                                recipe.reactants[idx].item.uid
                            )
                            for idx, _ in enumerate(recipe.reactants)
                        ]
                    )
                )
            )
            products_list = sorted(
                list(
                    set(
                        [
                            canonical_smiles_parsing(
                                recipe.products[idx].item.uid
                            )
                            for idx, _ in enumerate(recipe.products)
                        ]
                    )
                )
            )

            if products_list == reactant_list:
                return False

        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(
            operator_keys={"species_similarity_check"}
        )


@typing.final
@dataclasses.dataclass(frozen=True)
class CheckSpeciesMoeity(metadata.ReactionFilterBase):
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        if recipe.operator.meta["species_moeity_check"] is not None:
            for idx, _mol in enumerate(recipe.products):
                for moeity in recipe.operator.meta["species_moeity_check"]:
                    product_smi = canonical_smiles_parsing(
                        recipe.products[idx].item.uid
                    )
                    if (
                        Chem.MolFromSmiles(product_smi).HasSubstructMatch(
                            Chem.MolFromSmarts(moeity)
                        )
                        is True
                    ):
                        return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(operator_keys={"species_moeity_check"})


MAX_C_ATOM = 10
MAX_O_ATOM = 2


@dataclasses.dataclass(frozen=True, slots=True)
class custom_operator_rxn_smarts:
    name: str = None
    rxn_smarts: str = None
    radical_tuple: Optional[tuple[int, ...]] = None
    generation_tuple: Optional[tuple[int, ...]] = None
    oxygen_presence_tuple: Optional[Tuple[Union[bool, None]]] = None
    C_atom_check: Optional[Union[bool, None]] = None
    max_permit_carbon: Optional[Tuple[Union[bool, None]]] = None
    OXY_ABSTRACT_COUNT: Optional[
        Tuple[Union[bool, int, float, None]] | Union[bool, int, float, None]
    ] = None
    species_similarity_check: Optional[bool] = None
    species_moeity_check: Optional[str] = None
    id_position_check: Optional[Union[str, bool, None]] = None
    substrate_selection_check: Optional[
        Union[bool, None] | Tuple[Union[bool, None]]
    ] = None


operator_smarts = (
    # 1 Primary Initation
    custom_operator_rxn_smarts(
        name="oxyinit",  # RH + O2 --> R. + HOO.
        rxn_smarts="[C:1][H:2].[O+0:3]=[O+0:4]>>[*H0:1].[H:2][*:3][*H0:4]",
        radical_tuple=(0, 0),
        C_atom_check=True,
        max_permit_carbon=(MAX_C_ATOM, None),
        substrate_selection_check=(True, None),
    ),
    # 2 Bond Fission
    custom_operator_rxn_smarts(
        name="init",
        rxn_smarts="[C+0:2][O:3][O+0H:4]>>[*:2][*H0:3].[*H0:4]",
        radical_tuple=(0,),
        oxygen_presence_tuple=(True,),
        max_permit_carbon=(MAX_C_ATOM,),
    ),
    # 2 Hydroperoxide Decomposition
    custom_operator_rxn_smarts(
        name="hpdecomp",
        rxn_smarts="[C:1][O:2][OX2+0H:3].[C:4][H:5]>>[*:1][*H0:2].[*:3][H:5].[*H0:4]",
        radical_tuple=(0, 0),
        C_atom_check=False,
        substrate_selection_check=(None, True),
    ),
    # 4 Oxygen Addition
    custom_operator_rxn_smarts(
        name="oxygadd",
        rxn_smarts="[Cv3:1].[O:2]=[O:3]>>[*:1][*:2][*H0:3]",
        radical_tuple=(1, 0),
    ),
    # Type: beta_scission_from_alkoxy_rxn
    custom_operator_rxn_smarts(
        name="oxybscission",
        rxn_smarts="[OX1:1][C:2][C+0:3]>>[*:1]=[*:2].[*H0:3]",
        radical_tuple=(1,),
        oxygen_presence_tuple=(True,),
    ),
    custom_operator_rxn_smarts(
        name="oxybscission",
        rxn_smarts="[C+0:1][O:2][OX1:3]>>[*:2]=[*:3].[*H0:1]",
        radical_tuple=(1,),
        oxygen_presence_tuple=(True,),
    ),
    # beta_scission_from_C_centered_radical_rxn
    custom_operator_rxn_smarts(
        name="bscission",
        rxn_smarts="[Cv3:1][O:2][O+0H:3]>>[*:1]=[*:2].[*H0:3]",
        radical_tuple=(1,),
        oxygen_presence_tuple=(True,),
    ),
    custom_operator_rxn_smarts(
        name="bscission",
        rxn_smarts="[Cv3:1][O:2][OX2+0:3]>>[*:1]=[*:2].[*H0:3]",
        radical_tuple=(1,),
        oxygen_presence_tuple=(True,),
    ),
    # hydrogen_transfer_habsOxy_rxn
    custom_operator_rxn_smarts(
        name="habsOxy",
        rxn_smarts="[H:5][C+0:1][H:2].[C:3][OX1:4]>>[*H0:1][H:2].[*:3][*:4][H:5]",
        radical_tuple=(0, 1),
        oxygen_presence_tuple=(None, True),
        OXY_ABSTRACT_COUNT=(MAX_O_ATOM, None),
        species_similarity_check=True,
    ),
    # hydrogen_transfer_alkylperoxy_radical_and_alkane_rxn
    custom_operator_rxn_smarts(
        name="habsPOxy",
        rxn_smarts="[H:6][C+0:1][H:2].[C:3][O:4][OX1:5]>>[*H0:1][H:2].[*:3][*:4][*:5][H:6]",
        radical_tuple=(0, 1),
        OXY_ABSTRACT_COUNT=(MAX_O_ATOM, None),
        species_similarity_check=True,
        species_moeity_check=[
            "[*:1][CH2:2][CH0:3]=[O:4]",
            "[CH0:1][O:2][OH1:3]",
            "[*:1][CH1:2][O:3][OH1:4]",
            "[*:1][CH1:1][C:2](=[O:3])[CH2:4][*:5]",
        ],
    ),
    # hydrogen_transfer_alkylperoxy_radical_and_aldehydes_rxn
    custom_operator_rxn_smarts(
        name="habsPCOxy",
        rxn_smarts="[C:1][O+0:2][O+0X1:3].[CX3+0:4]([H:6])=[O+0:5]>>[*:1][*:2][*:3][H:6].[*H0:4]=[*:5]",
        radical_tuple=(1, 0),
        oxygen_presence_tuple=(True, True),
        OXY_ABSTRACT_COUNT=(None, MAX_O_ATOM),
        species_similarity_check=True,
    ),
    # hydrogen_transfer_alkylperoxy_radical_and_alkyl_hydroperoxides_rxn
    custom_operator_rxn_smarts(
        name="habsKET",
        rxn_smarts="[C+0:2]([H:8])[O:3][O+0H:4].[C:5][O:6][OX1H0:7]>>[*H0:2][*:3][*:4].[*:5][*:6][*:7][*:8]",
        radical_tuple=(0, 1),
        oxygen_presence_tuple=(True, True),
        OXY_ABSTRACT_COUNT=(MAX_O_ATOM, None),
        species_similarity_check=True,
    ),
    # hydrogen transfer adjacent to a CO group in ketones
    custom_operator_rxn_smarts(
        name="habsKET2",
        rxn_smarts="[C+0:1]([H:7])[C:2](=[O:3])[C:8].[C:4][O:5][OX1H0:6]>>[*H0:1][*:2](=[*:3])[*:8].[*:4][*:5][*:6][H:7]",
        radical_tuple=(0, 1),
        oxygen_presence_tuple=(True, True),
        max_permit_carbon=(None, None),
        OXY_ABSTRACT_COUNT=(MAX_O_ATOM, None),
        species_similarity_check=True,
        species_moeity_check=["CCC(=O)[CH2]", "C[CH]C=O"],
    ),
    custom_operator_rxn_smarts(
        name="habsKET2",
        rxn_smarts="[C+0:1]([H:6])[C:2](=[O:3])[C:7].[C:4][OX1H0:5]>>[*H0:1][*:2](=[*:3])[*:7].[*:4][*:5][H:6]",
        radical_tuple=(0, 1),
        oxygen_presence_tuple=(True, True),
        OXY_ABSTRACT_COUNT=(MAX_O_ATOM, None),
        species_similarity_check=True,
        species_moeity_check=["CCC(=O)[CH2]", "C[CH]C=O"],
    ),
    # Other types of H-abstraction
    custom_operator_rxn_smarts(
        name="habs",
        rxn_smarts="[C,O,H0:1][O+0:2][H:5].[C,O,H0:3][OX1:4]>>[*:1][*H0:2].[*:3][*:4][H:5]",
        radical_tuple=(0, 1),
        oxygen_presence_tuple=(None, True),
        OXY_ABSTRACT_COUNT=(MAX_O_ATOM, None),
        species_similarity_check=True,
    ),
    custom_operator_rxn_smarts(
        name="habs",
        rxn_smarts="[#6;H1,H2,O+0:1][H:2].[O+0H,!O;!C:3][Ov1+0:4]>>[*H0:1].[*:3][*:4][H:2]",
        radical_tuple=(0, 1),
        oxygen_presence_tuple=(None, True),
        OXY_ABSTRACT_COUNT=(MAX_O_ATOM, None),
        species_similarity_check=True,
    ),
    # Disproportionation (Type A)
    custom_operator_rxn_smarts(
        name="perdispropcross",
        rxn_smarts="[C:1]([H:7])[O:2][OX1:3].[C:4][O:5][OX1:6]>>[*:4][*:5][H:7].[*:1]=[*:2].[O:3]=[O:6]",
        radical_tuple=(1, 1),
        oxygen_presence_tuple=(True, True),
        id_position_check=None,
    ),
    custom_operator_rxn_smarts(
        name="perdispropcross",
        rxn_smarts="[C:1]([H:6])[O:2][OX1:3].[Ov1+0:4][O+0H:5]>>[*:5][H:6].[*:1]=[*:2].[O:3]=[O:4]",
        radical_tuple=(1, 1),
        oxygen_presence_tuple=(True, True),
    ),
    # Disproportionation (Type B) type 1
    custom_operator_rxn_smarts(
        name="perdisprop2ho2",
        rxn_smarts="[O+0H:1][Ov1+0:2].[Ov1+0:3][O+0H:4]>>[*H0:1].[O:2]=[O:3].[*H0:4]",
        radical_tuple=(1, 1),
        oxygen_presence_tuple=(True, True),
        OXY_ABSTRACT_COUNT=None,
    ),
    custom_operator_rxn_smarts(
        name="perdisprop2ho2",
        rxn_smarts="[C:1][O:2][OX1:3].[Ov1+0:4][O+0H:5]>>[*:1][*H0:2].[*:3]=[*:4].[*H0:5]",
        radical_tuple=(1, 1),
        oxygen_presence_tuple=(True, True),
        OXY_ABSTRACT_COUNT=None,
    ),
    # Disproportionation (Type B) type 2
    custom_operator_rxn_smarts(
        name="perdisprop2",
        rxn_smarts="[C:1][O:2][OX1:3].[C:4][O:5][OX1:6]>>[*:1][*H0:2].[O:3]=[O:6].[*:4][*H0:5]",
        radical_tuple=(1, 1),
        oxygen_presence_tuple=(True, True),
        id_position_check=None,
        OXY_ABSTRACT_COUNT=None,
    ),
    # Recombination
    custom_operator_rxn_smarts(
        name="recomb",
        rxn_smarts="[C,O,H0:1][OX1:2].[C,O,H0:3][OX1:4]>>[*:1][*:2][*:4][*:3]",
        radical_tuple=(1, 1),
        max_permit_carbon=(MAX_C_ATOM, MAX_C_ATOM),
        species_moeity_check=[
            "COOOC",
            "COOO",
            "OOO",
            "COOOOC",
            "COOOO",
            "OOOO",
        ],
        id_position_check=None,
    ),
    custom_operator_rxn_smarts(
        name="recomb",
        rxn_smarts="[C,O,H0:1][OX1:2].[Cv3:3]>>[*:1][*:2][*:3]",
        radical_tuple=(1, 1),
        max_permit_carbon=(MAX_C_ATOM, MAX_C_ATOM),
        species_moeity_check=[
            "COOOC",
            "COOO",
            "OOO",
            "COOOOC",
            "COOOO",
            "OOOO",
        ],
    ),
    custom_operator_rxn_smarts(
        name="recomb",
        rxn_smarts="[Cv3:1].[Cv3:2]>>[*:1][*:2]",
        radical_tuple=(1, 1),
        max_permit_carbon=(MAX_C_ATOM, MAX_C_ATOM),
        id_position_check=None,
    ),
    custom_operator_rxn_smarts(
        name="recomb",
        rxn_smarts="[Cv3:1]=[O:2].[O+0H,!O;!C:3][Ov1+0:4]>>[*:1](=[O:2])[*:4][*:3]",
        radical_tuple=(1, 1),
        max_permit_carbon=(MAX_C_ATOM, None),
    ),
    # 17 Bayer-Villiger
    custom_operator_rxn_smarts(
        name="BV",
        rxn_smarts="[CX3+0:1]([H:7])=[O+0:2].[CX3:3](=[O+0:4])[O:5][O+0H:6]>>[*:1](=[O:2])[*:5].[*:3](=[O:4])[*:6]",
        radical_tuple=(0, 0),
        oxygen_presence_tuple=(True, True),
        C_atom_check=True,
    ),
)


initial_reactant_smiles = [
    update_isotope_smiles("CCCC"),
    update_isotope_smiles("O=O"),
]

for smiles in initial_reactant_smiles:
    network.add_mol(
        engine.mol.rdkit(smiles),
        meta={
            "gen": 0,
            "radical_num": Descriptors.NumRadicalElectrons(
                Chem.MolFromSmiles(canonical_smiles_parsing(smiles))
            ),
            "oxygen_presence": "O"
            in [
                atom.GetSymbol()
                for atom in Chem.MolFromSmiles(smiles).GetAtoms()
            ],
        },
    )

for smarts in operator_smarts:
    network.add_op(
        engine.op.rdkit(smarts.rxn_smarts),
        meta={
            "name": smarts.name,
            "rxn_smarts": smarts.rxn_smarts,
            "radical_tuple": smarts.radical_tuple,
            "generation_tuple": smarts.generation_tuple,
            "oxygen_presence_tuple": smarts.oxygen_presence_tuple,
            "C_atom_check": smarts.C_atom_check,
            "max_permit_carbon": smarts.max_permit_carbon,
            "OXY_ABSTRACT_COUNT": smarts.OXY_ABSTRACT_COUNT,
            "species_similarity_check": smarts.species_similarity_check,
            "species_moeity_check": smarts.species_moeity_check,
            "id_position_check": smarts.id_position_check,
        },
    )

max_iteration = 2

network.save_to_file("dummy_test")
network = engine.network_from_file("dummy_test")
strat = engine.strat.cartesian(network)
final_rank = 1
atom_num = [atom.GetAtomicNum() for atom in Chem.MolFromSmiles("C").GetAtoms()]

gen_filter = engine.filter.reaction.generation(final_rank, "gen")
max_atoms_filter = engine.filter.reaction.max_atoms(
    max_atoms=MAX_C_ATOM, proton_number=atom_num[0]
)

reaction_plan = (
    engine.meta.generation("gen")  # generation (i.e., rank)
)

strat.expand(
    num_iter=max_iteration,
    reaction_plan=reaction_plan,
    save_unreactive=False,
)

print("wow")



"""Synthetic network generation code."""

import collections.abc
import dataclasses
import re
import time
import typing
from datetime import datetime

from rdkit import Chem
from rdkit.Chem import rdqueries
from rdkit.Chem.rdMolDescriptors import CalcMolFormula

import doranet as dn
from doranet import interfaces, metadata
from doranet.modules.synthetic.Reaction_Smarts_Forward import op_smarts
from doranet.modules.synthetic.Reaction_Smarts_Retro import op_retro_smarts


@typing.final
@dataclasses.dataclass(frozen=True, slots=True)
class Chem_Rxn_dH_Calculator(metadata.RxnPropertyCalc[float]):
    dH_key: collections.abc.Hashable
    direction: str
    Mole_Hf: typing.Callable[[str], float]

    @property
    def key(self) -> collections.abc.Hashable:
        return self.dH_key

    @property
    def resolver(self) -> metadata.MetaDataResolverFunc[float]:
        return metadata.TrivialMetaDataResolverFunc

    def __call__(
        self,
        data: interfaces.ReactionExplicit,
        prev_value: typing.Optional[float] = None,
    ) -> typing.Optional[float]:
        if prev_value is not None:
            return prev_value
        if data.reaction_meta is not None and self.dH_key in data.reaction_meta:
            return None
        if self.Mole_Hf is None:
            return "No_Thermo"
        dH = 0.0

        if data.operator.meta is None:
            raise RuntimeError("No operator metadata found!")

        for idx, mol in enumerate(data.products):
            if not isinstance(mol.item, interfaces.MolDatRDKit):
                raise NotImplementedError(
                    f"""Calculator only implemented for molecule type \
                        MolDatRDKit, not {type(mol.item)}"""
                )
            pro_Hf = self.Mole_Hf(mol.item.smiles)
            if pro_Hf is None:
                return float("nan")
            dH = dH + pro_Hf * data.operator.meta["products_stoi"][idx]
        for idx, mol in enumerate(data.reactants):
            if not isinstance(mol.item, interfaces.MolDatRDKit):
                raise NotImplementedError(
                    f"""Calculator only implemented for molecule type \
                        MolDatRDKit, not {type(mol.item)}"""
                )
            rea_Hf = self.Mole_Hf(mol.item.smiles)
            if rea_Hf is None:
                return float("nan")
            dH = dH - rea_Hf * data.operator.meta["reactants_stoi"][idx]
        if data.operator.meta["enthalpy_correction"] is not None:
            dH = dH + data.operator.meta["enthalpy_correction"]
        if self.direction == "forward":
            dH = dH / data.operator.meta["number_of_steps"]
        if self.direction == "retro":
            dH = -dH / data.operator.meta["number_of_steps"]
        return round(dH, 4)

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(
            operator_keys={
                "reactants_stoi",
                "products_stoi",
                "enthalpy_correction",
                "number_of_steps",
            }
        )


@typing.final
@dataclasses.dataclass(frozen=True)
class Rxn_dH_Filter(metadata.ReactionFilterBase):
    __slots__ = ("max_dH", "dH_key")
    max_dH: float
    dH_key: collections.abc.Hashable

    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        # if metadata not available, reject reaction
        # (clearly something has gone wrong here)
        if recipe.reaction_meta is None:
            return False
        dH = recipe.reaction_meta[self.dH_key]
        if dH == "No_Thermo":
            return True
        if dH == float("nan"):
            return False
        return dH < self.max_dH

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket()


@typing.final
@dataclasses.dataclass(frozen=True)
class Ring_Issues_Filter(metadata.ReactionFilterBase):
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        if recipe.operator.meta is None:
            raise RuntimeError("No operator metadata found!")

        if (
            recipe.operator.meta["ring_issue"] is True
            and recipe.operator.meta["enthalpy_correction"] is None
        ):
            reactants_dict: dict[str, float | int] = dict()
            products_dict: dict[str, float | int] = dict()
            pattern = r"([A-Z][a-z]*)(\d*)"
            for idx, mol in enumerate(recipe.reactants):
                if not isinstance(mol.item, interfaces.MolDatRDKit):
                    raise NotImplementedError(
                        f"""Calculator only implemented for molecule type \
                            MolDatRDKit, not {type(mol.item)}"""
                    )
                smiles = CalcMolFormula(mol.item.rdkitmol)
                matches = re.findall(pattern, smiles)
                for match in matches:
                    element, count = match
                    count = int(count) if count else 1
                    reactants_dict[element] = (
                        reactants_dict.get(element, 0)
                        + count * recipe.operator.meta["reactants_stoi"][idx]
                    )
            for idx, mol in enumerate(recipe.products):
                if not isinstance(mol.item, interfaces.MolDatRDKit):
                    raise NotImplementedError(
                        f"""Calculator only implemented for molecule type \
                            MolDatRDKit, not {type(mol.item)}"""
                    )
                if "." in mol.item.smiles:
                    # if there're fragments in a mol, indicates invalid rxn
                    return False
                smiles = CalcMolFormula(mol.item.rdkitmol)
                matches = re.findall(pattern, smiles)
                for match in matches:
                    element, count = match
                    count = int(count) if count else 1
                    products_dict[element] = (
                        products_dict.get(element, 0)
                        + count * recipe.operator.meta["products_stoi"][idx]
                    )
            if reactants_dict != products_dict:
                return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(
            operator_keys={
                "reactants_stoi",
                "products_stoi",
                "ring_issue",
                "enthalpy_correction",
            }
        )


@typing.final
@dataclasses.dataclass(frozen=True)
class Retro_Not_Aromatic_Filter(metadata.ReactionFilterBase):
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        rea_aro_ring_num = 0
        pro_aro_ring_num = 0

        if recipe.operator.meta is None:
            raise RuntimeError("No operator metadata found!")

        if recipe.operator.meta["Retro_Not_Aromatic"] is True:
            for idx, mol in enumerate(recipe.reactants):
                if not isinstance(mol.item, interfaces.MolDatRDKit):
                    raise NotImplementedError(
                        f"""Filter only implemented for molecule type \
                            MolDatRDKit, not {type(mol.item)}"""
                    )
                rea_aro_ring_num += (
                    Chem.rdMolDescriptors.CalcNumAromaticRings(
                        mol.item.rdkitmol
                    )
                    * recipe.operator.meta["reactants_stoi"][idx]
                )
            for idx, mol in enumerate(recipe.products):
                if not isinstance(mol.item, interfaces.MolDatRDKit):
                    raise NotImplementedError(
                        f"""Filter only implemented for molecule type \
                            MolDatRDKit, not {type(mol.item)}"""
                    )
                pro_aro_ring_num += (
                    Chem.rdMolDescriptors.CalcNumAromaticRings(
                        mol.item.rdkitmol
                    )
                    * recipe.operator.meta["products_stoi"][idx]
                )
            if rea_aro_ring_num < pro_aro_ring_num:
                return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(
            operator_keys={
                "reactants_stoi",
                "products_stoi",
                "Retro_Not_Aromatic",
            }
        )


@typing.final
@dataclasses.dataclass(frozen=True)
class Check_balance_filter(metadata.ReactionFilterBase):
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        if (
            recipe.operator.meta is not None
            and recipe.operator.meta["enthalpy_correction"] is None
            and recipe.operator.meta["ring_issue"] is False
        ):
            reactants_dict: dict[str, float | int] = dict()
            products_dict: dict[str, float | int] = dict()
            pattern = r"([A-Z][a-z]*)(\d*)"
            for idx, mol in enumerate(recipe.reactants):
                if not isinstance(mol.item, interfaces.MolDatRDKit):
                    raise NotImplementedError(
                        f"""Filter only implemented for molecule type \
                            MolDatRDKit, not {type(mol.item)}"""
                    )
                smiles = CalcMolFormula(mol.item.rdkitmol)
                matches = re.findall(pattern, smiles)
                for match in matches:
                    element, count = match
                    count = int(count) if count else 1
                    reactants_dict[element] = (
                        reactants_dict.get(element, 0)
                        + count * recipe.operator.meta["reactants_stoi"][idx]
                    )
            for idx, mol in enumerate(recipe.products):
                if not isinstance(mol.item, interfaces.MolDatRDKit):
                    raise NotImplementedError(
                        f"""Filter only implemented for molecule type \
                            MolDatRDKit, not {type(mol.item)}"""
                    )
                smiles = CalcMolFormula(mol.item.rdkitmol)
                matches = re.findall(pattern, smiles)
                for match in matches:
                    element, count = match
                    count = int(count) if count else 1
                    products_dict[element] = (
                        products_dict.get(element, 0)
                        + count * recipe.operator.meta["products_stoi"][idx]
                    )
            if reactants_dict != products_dict:
                return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(
            operator_keys={
                "reactants_stoi",
                "products_stoi",
                "ring_issue",
                "enthalpy_correction",
                "name",
            }
        )


@typing.final
@dataclasses.dataclass(frozen=True)
class Cross_Reaction_Filter(interfaces.RecipeFilter):
    # no rea1 + rea2 or pure helper rxns. rea1 + rea1 or rea1 + helper ok
    # __slots__ = "coreactants"
    coreactants: collections.abc.Container[interfaces.MolIndex]

    def __call__(self, recipe: interfaces.RecipeExplicit) -> bool:
        reactant_set = set()
        for mol in recipe.reactants:
            if mol.i not in self.coreactants:
                reactant_set.add(mol.i)
        return len(set(reactant_set)) == 1

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket()


@typing.final
@dataclasses.dataclass(frozen=True)
class Enol_filter_forward(metadata.ReactionFilterBase):
    # if an enol is in reactants, only allow tautomerization rxn
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        for mol in recipe.reactants:
            if not isinstance(mol.item, interfaces.MolDatRDKit):
                raise NotImplementedError(
                    f"""Calculator only implemented for molecule type \
                        MolDatRDKit, not {type(mol.item)}"""
                )
            if (
                mol.item.rdkitmol.HasSubstructMatch(
                    Chem.MolFromSmarts("[C]=[C]-[OH]")
                )
                and recipe.operator.meta is not None
                and (
                    recipe.operator.meta["name"]
                    != "Keto-enol Tautomerization Reverse"
                )
            ):
                return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(operator_keys={"name"})


@typing.final
@dataclasses.dataclass(frozen=True)
class Enol_filter_retro(metadata.ReactionFilterBase):
    # if an enol is in reactants, only allow tautomerization rxn
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        for mol in recipe.products:
            if not isinstance(mol.item, interfaces.MolDatRDKit):
                raise NotImplementedError(
                    f"""Filter only implemented for molecule type \
                        MolDatRDKit, not {type(mol.item)}"""
                )
            if (
                mol.item.rdkitmol.HasSubstructMatch(
                    Chem.MolFromSmarts("[C]=[C]-[OH]")
                )
                and recipe.operator.meta is not None
                and (
                    recipe.operator.meta["name"]
                    != "Keto-enol Tautomerization Reverse"
                )
            ):
                return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(operator_keys={"name"})


@typing.final
@dataclasses.dataclass(frozen=True)
class Allowed_Elements_Filter(metadata.ReactionFilterBase):
    # only allow reactions with specified elements in reactants.
    # does not check hydrogen
    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        if (
            recipe.operator.meta is None
            or recipe.operator.meta["allowed_elements"][0] == "All"
        ):
            return True
        for mol in recipe.reactants:
            if not isinstance(mol.item, interfaces.MolDatRDKit):
                raise NotImplementedError(
                    f"""Filter only implemented for molecule type \
                        MolDatRDKit, not {type(mol.item)}"""
                )
            for atom in mol.item.rdkitmol.GetAtoms():
                if (
                    atom.GetSymbol()
                    not in recipe.operator.meta["allowed_elements"]
                ):
                    return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket(operator_keys={"allowed_elements"})


@typing.final
@dataclasses.dataclass(frozen=True, slots=True)
class Max_Atoms_Filter(metadata.ReactionFilterBase):
    # similar to engine.filter.reaction.max_atoms
    # take a dict, check multiple atoms {6: 12, 8: 2}
    max_atoms_dict: dict

    def __call__(self, recipe: interfaces.ReactionExplicit) -> bool:
        if self.max_atoms_dict is None:
            return True
        for atom_number, max_atoms in self.max_atoms_dict.items():
            for mol in recipe.products:
                if not isinstance(mol.item, interfaces.MolDatRDKit):
                    raise NotImplementedError(
                        f"""Filter only implemented for molecule type \
                            MolDatRDKit, not {type(mol.item)}"""
                    )
                atom_q = rdqueries.AtomNumEqualsQueryAtom(atom_number)
                atom_matches = mol.item.rdkitmol.GetAtomsMatchingQuery(atom_q)
                if len(atom_matches) > max_atoms:
                    return False
        return True

    @property
    def meta_required(self) -> interfaces.MetaKeyPacket:
        return interfaces.MetaKeyPacket()


def get_smiles_from_file(file_name):
    def is_valid_smiles(smiles_string):
        try:
            mol = Chem.MolFromSmiles(smiles_string)
            return mol is not None
        except ValueError:
            return False

    if not isinstance(file_name, str):
        return file_name
    if is_valid_smiles(file_name):
        return [file_name]
    with open(
        file_name  # , encoding="utf-8"
    ) as f:
        lines = f.readlines()
    clean_list = list()
    for i in lines:
        if i != "\n":
            clean_list.append(i.strip())
    return clean_list


def generate_network(
    job_name="default_job",
    starters=False,
    helpers=False,
    gen=1,
    direction="forward",
    molecule_thermo_calculator=None,
    max_rxn_thermo_change=15,
    max_atoms=None,  # {"C": 20}
    allow_multiple_reactants="default",  # forward allowed, retro no
    targets=None,  # string or list, set, etc.
):
    if not starters:
        raise Exception("At least one starter is needed to generate a network")

    starters = get_smiles_from_file(starters)
    helpers = get_smiles_from_file(helpers)
    targets = get_smiles_from_file(targets)

    print(f"Job name: {job_name}")
    print(f"Job type: synthetic network expansion {direction}")
    print("Job started on:", datetime.now())
    start_time = time.time()

    engine = dn.create_engine()
    network = engine.new_network()

    if helpers:
        for smiles in helpers:
            network.add_mol(engine.mol.rdkit(smiles))

    my_start_i = -1
    for smiles in starters:
        if my_start_i == -1:
            my_start_i = network.add_mol(engine.mol.rdkit(smiles))
        else:
            network.add_mol(engine.mol.rdkit(smiles))

    if direction == "forward":
        smarts_list = op_smarts
    elif direction == "retro":
        smarts_list = op_retro_smarts

    for smarts in smarts_list:
        if smarts.kekulize_flag is False:
            network.add_op(
                engine.op.rdkit(
                    smarts.smarts,
                    drop_errors=True,
                ),
                meta={
                    "name": smarts.name,
                    "reactants_stoi": smarts.reactants_stoi,
                    "products_stoi": smarts.products_stoi,
                    "enthalpy_correction": smarts.enthalpy_correction,
                    "ring_issue": smarts.ring_issue,
                    "kekulize_flag": smarts.kekulize_flag,
                    "Retro_Not_Aromatic": smarts.Retro_Not_Aromatic,
                    "number_of_steps": smarts.number_of_steps,
                    "allowed_elements": smarts.allowed_elements,
                    "Reaction_type": smarts.reaction_type,
                    "Reaction_direction": direction,
                },
            )
        if smarts.kekulize_flag is True:
            network.add_op(
                engine.op.rdkit(
                    smarts.smarts,
                    kekulize=True,
                    drop_errors=True,
                ),
                meta={
                    "name": smarts.name,
                    "reactants_stoi": smarts.reactants_stoi,
                    "products_stoi": smarts.products_stoi,
                    "enthalpy_correction": smarts.enthalpy_correction,
                    "ring_issue": smarts.ring_issue,
                    "kekulize_flag": smarts.kekulize_flag,
                    "Retro_Not_Aromatic": smarts.Retro_Not_Aromatic,
                    "number_of_steps": smarts.number_of_steps,
                    "allowed_elements": smarts.allowed_elements,
                    "Reaction_type": smarts.reaction_type,
                    "Reaction_direction": direction,
                },
            )

    strat = engine.strat.cartesian(network)

    periodic_table = Chem.GetPeriodicTable()

    if max_atoms is None:
        max_atoms_dict_num = None
    else:
        max_atoms_dict_num = dict()
        for key in max_atoms:
            max_atoms_dict_num[periodic_table.GetAtomicNumber(key)] = max_atoms[
                key
            ]

    if direction == "forward":
        reaction_plan = (
            Max_Atoms_Filter(max_atoms_dict_num)
            >> Ring_Issues_Filter()
            >> Enol_filter_forward()
            >> Check_balance_filter()
            >> Allowed_Elements_Filter()
            >> Chem_Rxn_dH_Calculator(
                "dH", "forward", molecule_thermo_calculator
            )
            >> Rxn_dH_Filter(max_rxn_thermo_change, "dH")
        )
        recipe_filter = None

    elif direction == "retro":
        reaction_plan = (
            Max_Atoms_Filter(max_atoms_dict_num)
            >> Ring_Issues_Filter()
            >> Retro_Not_Aromatic_Filter()
            >> Enol_filter_retro()
            >> Allowed_Elements_Filter()
            >> Check_balance_filter()
            >> Chem_Rxn_dH_Calculator("dH", "retro", molecule_thermo_calculator)
            >> Rxn_dH_Filter(max_rxn_thermo_change, "dH")
        )
        recipe_filter = Cross_Reaction_Filter(tuple(range(my_start_i)))

    if allow_multiple_reactants != "default":
        if allow_multiple_reactants is True:
            recipe_filter = None
        elif allow_multiple_reactants is False:
            recipe_filter = Cross_Reaction_Filter(tuple(range(my_start_i)))

    bundle_filter = engine.filter.bundle.coreactants(tuple(range(my_start_i)))

    ini_number = len(network.mols)

    strat.expand(
        num_iter=gen,
        reaction_plan=reaction_plan,
        bundle_filter=bundle_filter,
        recipe_filter=recipe_filter,
        save_unreactive=False,
    )

    if targets is not None:
        print("Checking for targets")
        to_check = set()
        if isinstance(targets, str):
            to_check.add(Chem.MolToSmiles(Chem.MolFromSmiles(targets)))
        else:
            for i in targets:
                to_check.add(Chem.MolToSmiles(Chem.MolFromSmiles(i)))

        for mol in network.mols:
            if (
                network.reactivity[network.mols.i(mol.uid)] is True
                and mol.uid in to_check
            ):
                print("Target found for", mol.uid)

    print("Number of generations:", gen)
    print("Number of operators:", len(network.ops))
    print("Number of molecules before expantion:", ini_number)
    print("Number of molecules after expantion:", len(network.mols))
    print("Number of reactions:", len(network.rxns))

    end_time = time.time()
    elapsed_time = (end_time - start_time) / 60
    print(
        "Time used for network generation:",
        "{:.2f}".format(elapsed_time),
        "minutes",
    )
    print()

    network.save_to_file(f"{job_name}_{direction}_saved_network")

    return network



"""Test tutorial exercises."""

import os
import tempfile

import doranet as dn


def test_tutorial_2():
    engine = dn.create_engine()
    acetone = engine.mol.rdkit("CC(=O)C")

    assert repr(acetone) == "MolDatBasic('CC(C)=O')"
    assert acetone.uid == "CC(C)=O"
    assert acetone.smiles == "CC(C)=O"

    aldol_condensation = engine.op.rdkit(
        "[O&+0:1]=[C&+0:2]-[C&+0;H2,H3:3].[C&+0:4]=[O&+0:5]>>[*:1]=[*:2]-[*:3]=[*:4].[*:5]"
    )

    assert (
        repr(aldol_condensation(acetone, acetone))
        == "((MolDatBasic('CC(=O)C=C(C)C'), MolDatBasic('O')), (MolDatBasic('CC(=O)C=C(C)C'), MolDatBasic('O')))"
    )


def test_tutorial_3():
    engine = dn.create_engine()

    water = engine.mol.rdkit("O")
    ethanol = engine.mol.rdkit("CCO")
    acetone = engine.mol.rdkit("CC(C)=O")
    butanone = engine.mol.rdkit("CCC(C)=O")
    methyl_butanoate = engine.mol.rdkit("CCCC(=O)OC")
    delta_valerolactone = engine.mol.rdkit("O=C1CCCCO1")
    hydroxyvaleric_acid = engine.mol.rdkit("O=C(O)CCCCO")

    aldol_condensation = engine.op.rdkit(
        "[O&+0:1]=[C&+0:2]-[C&+0;H2,H3:3].[C&+0:4]=[O&+0:5]>>[*:1]=[*:2]-[*:3]=[*:4].[*:5]"
    )

    assert (
        repr(aldol_condensation(acetone, acetone))
        == "((MolDatBasic('CC(=O)C=C(C)C'), MolDatBasic('O')), (MolDatBasic('CC(=O)C=C(C)C'), MolDatBasic('O')))"
    )

    assert (
        repr(aldol_condensation(butanone, acetone))
        == "((MolDatBasic('CC(=O)C(C)=C(C)C'), MolDatBasic('O')), (MolDatBasic('CCC(=O)C=C(C)C'), MolDatBasic('O')))"
    )

    assert aldol_condensation.compat(acetone, 0) is True

    assert aldol_condensation.compat(water, 1) is False

    assert aldol_condensation(acetone, water) == tuple()

    assert (
        aldol_condensation.smarts
        == "[O&+0:1]=[C&+0:2]-[C&+0;H2,H3:3].[C&+0:4]=[O&+0:5]>>[*:1]=[*:2]-[*:3]=[*:4].[*:5]"
    )

    assert len(aldol_condensation) == 2  # noqa: PLR2004

    try:
        aldol_condensation(acetone)
        raise AssertionError()
    except RuntimeError:
        assert True

    esterification = engine.op.rdkit(
        "[O&+0:1]=[C&+0:2]-[O&+0&H1:3].[O&+0&H1:4]>>[*:1]=[*:2]-[*:4].[*:3]"
    )

    assert (
        repr(esterification(hydroxyvaleric_acid, ethanol))
        == "((MolDatBasic('CCOC(=O)CCCCO'), MolDatBasic('O')),)"
    )

    esterification_intra = engine.op.rdkit(
        "([O&+0:1]=[C&+0:2]-[O&+0&H1:3].[O&+0&H1:4])>>[*:1]=[*:2]-[*:4].[*:3]"
    )

    assert (
        repr(esterification_intra(hydroxyvaleric_acid))
        == "((MolDatBasic('O=C1CCCCO1'), MolDatBasic('O')),)"
    )

    try:
        esterification_intra(hydroxyvaleric_acid, ethanol)
        raise AssertionError()
    except RuntimeError:
        assert True

    ester_hydrolysis_incorrect = engine.op.rdkit(
        "[O&+0:1]=[C&+0:2]-[O&+0&H0:3].[O&+0&H2:4]>>[*:1]=[*:2]-[*:4].[*:3]"
    )

    assert (
        repr(ester_hydrolysis_incorrect(methyl_butanoate, water))
        == "((MolDatBasic('CCCC(=O)O'), MolDatBasic('CO')),)"
    )

    assert (
        repr(ester_hydrolysis_incorrect(delta_valerolactone, water))
        == "((MolDatBasic('CCCCC(=O)O'), MolDatBasic('CCCCO')),)"
    )

    ester_hydrolysis_nonring = engine.op.rdkit(
        "[O&+0:1]=[C&+0:2]-&!@[O&+0&H0:3].[O&+0&H2:4]>>[*:1]=[*:2]-[*:4].[*:3]"
    )
    ester_hydrolysis_ring = engine.op.rdkit(
        "[O&+0:1]=[C&+0:2]-&@[O&+0&H0:3].[O&+0&H2:4]>>([*:1]=[*:2]-[*:4].[*:3])"
    )

    assert (
        repr(ester_hydrolysis_nonring(methyl_butanoate, water))
        == "((MolDatBasic('CCCC(=O)O'), MolDatBasic('CO')),)"
    )

    assert ester_hydrolysis_nonring(delta_valerolactone, water) == tuple()

    assert ester_hydrolysis_ring(methyl_butanoate, water) == tuple()

    assert (
        repr(ester_hydrolysis_ring(delta_valerolactone, water))
        == "((MolDatBasic('O=C(O)CCCCO'),),)"
    )


def test_tutorial_4():
    engine = dn.create_engine()

    water = engine.mol.rdkit("O")
    ethanol = engine.mol.rdkit("CCO")
    butanone = engine.mol.rdkit("CCC(C)=O")
    delta_valerolactone = engine.mol.rdkit("O=C1CCCCO1")
    hydroxyvaleric_acid = engine.mol.rdkit("O=C(O)CCCCO")

    ester_hydrolysis_ring = engine.op.rdkit(
        "[O&+0:1]=[C&+0:2]-&@[O&+0&H0:3].[O&+0&H2:4]>>([*:1]=[*:2]-[*:4].[*:3])"
    )

    network = engine.new_network()

    water_i = network.add_mol(water)

    assert water_i == 0

    assert repr(network.mols[water_i]) == "MolDatBasic('O')"

    assert water in network.mols
    assert water.uid in network.mols
    assert butanone not in network.mols

    assert network.mols.i(water.uid) == 0

    assert repr(network.mols[water_i]) == "MolDatBasic('O')"

    assert network.add_mol(water) == 0

    assert len(network.mols) == 1

    ester_hydrolysis_ring_i = network.add_op(ester_hydrolysis_ring)

    assert ester_hydrolysis_ring_i == 0

    network.add_mol(delta_valerolactone)

    hydroxyvaleric_acid = ester_hydrolysis_ring(delta_valerolactone, water)[0][
        0
    ]

    assert repr(hydroxyvaleric_acid) == "MolDatBasic('O=C(O)CCCCO')"

    water_i = network.mols.i(water.uid)
    delta_valerolactone_i = network.mols.i(delta_valerolactone.uid)
    ester_hydrolysis_ring_i = network.ops.i(ester_hydrolysis_ring.uid)

    hydroxyvaleric_acid = ester_hydrolysis_ring(delta_valerolactone, water)[0][
        0
    ]
    hydroxyvaleric_acid_i = network.add_mol(hydroxyvaleric_acid)

    reaction_i = network.add_rxn(
        operator=ester_hydrolysis_ring_i,
        reactants=(delta_valerolactone_i, water_i),
        products=(hydroxyvaleric_acid_i,),
    )

    assert reaction_i == 0
    assert network.rxns[0] == dn.interfaces.Reaction(0, (1, 0), (2,))

    assert network.consumers(0) == [0]
    assert network.producers(0) == []
    assert network.producers(2) == [0]

    with tempfile.TemporaryDirectory() as temp_dir:
        network.save_to_file("saved_network", temp_dir)
        assert os.listdir(temp_dir) == ["saved_network.pgnet"]

        network_loaded = engine.network_from_file("saved_network", temp_dir)
        assert (
            repr(list(network_loaded.mols))
            == "[MolDatBasic('O'), MolDatBasic('O=C1CCCCO1'), MolDatBasic('O=C(O)CCCCO')]"
        )

    assert network.compat_table(ester_hydrolysis_ring_i) == ([1], [0])

    network.add_mol(water, reactive=False)

    ethanol_i = network.add_mol(ethanol, meta={"is_alcohol": True})
    water_i = network.add_mol(
        water, meta={"is_alcohol": False, "solvent": True}
    )

    network.mols.set_meta(water_i, {"is_water": True, "is_toxic": False})

    assert network.mols.meta(water_i) == {
        "is_alcohol": False,
        "solvent": True,
        "is_water": True,
        "is_toxic": False,
    }
    assert network.mols.meta(water_i, ["is_alcohol"]) == {"is_alcohol": False}
    assert network.mols.meta([ethanol_i], ["is_alcohol"]) == (
        {"is_alcohol": True},
    )
    assert network.mols.meta([water_i, ethanol_i], keys=["is_alcohol"]) == (
        {"is_alcohol": False},
        {"is_alcohol": True},
    )
    assert network.mols.meta([ethanol_i, water_i], keys=["solvent"]) == (
        {},
        {"solvent": True},
    )
    assert network.mols.meta(keys=["is_alcohol"]) == (
        {"is_alcohol": False},
        {},
        {},
        {"is_alcohol": True},
    )
    assert network.mols.meta() == (
        {
            "is_alcohol": False,
            "solvent": True,
            "is_water": True,
            "is_toxic": False,
        },
        {},
        {},
        {"is_alcohol": True},
    )



import doranet

engine = doranet.create_engine()
network = engine.new_network()
network.add_mol(engine.mol.rdkit("C#C"), meta={"gen": 0})
network.add_op(engine.op.rdkit("[C:1]#[C:2]>>[*:1]=[*:2]"))
network.add_op(engine.op.rdkit("[C:1]#[C:2]>>[*:1]-[*:2]"))
network.add_op(engine.op.rdkit("[C:1]=[C:2]>>[*:1]-[*:2]"))

strat = engine.strat.cartesian(network)

strat.expand(reaction_plan=doranet.metacalc.GenerationCalculator("gen"))

print("testo")



"""A streamlined, generic rewrite of Pickaxe."""

__all__ = [
    "create_engine",
    "datatypes",
    "engine",
    "filters",
    "hooks",
    "interfaces",
    "metacalc",
    "metadata",
    "modules",
    "network",
    "strategies",
    "utils",
]

from doranet.engine import create_engine

from . import (
    datatypes,
    engine,
    filters,
    hooks,
    interfaces,
    metacalc,
    metadata,
    modules,
    network,
    strategies,
    utils,
)



"""Contains classes which define and implement network expansion strategies."""

import collections.abc
import dataclasses
import heapq
import itertools
import math
import typing

from doranet import interfaces, metadata
from doranet import network as pgnetworks


def _generate_recipes_from_compat_table(
    compat: dict[interfaces.Identifier, list[list[interfaces.Identifier]]],
    known_cache: typing.Optional[
        set[tuple[interfaces.Identifier, tuple[interfaces.Identifier, ...]]]
    ] = None,
    uid_prefilter: typing.Optional[
        collections.abc.Callable[
            [
                interfaces.Identifier,
                collections.abc.Sequence[interfaces.Identifier],
            ],
            bool,
        ]
    ] = None,
    add_to_cache: bool = True,
) -> collections.abc.Generator[
    tuple[interfaces.Identifier, tuple[interfaces.Identifier, ...]], None, None
]:
    gen_cache = set()
    for op_uid, compat_list in compat.items():
        reactants: tuple[interfaces.Identifier, ...]
        for reactants in itertools.product(*compat_list):
            recipe = (op_uid, reactants)
            if known_cache is not None:
                if recipe in known_cache:
                    continue
                elif add_to_cache:
                    known_cache.add(recipe)
            if uid_prefilter is not None and not uid_prefilter(
                op_uid, reactants
            ):
                continue
            recipe = (op_uid, reactants)
            if recipe not in gen_cache:
                gen_cache.add(recipe)
                yield (op_uid, reactants)
            else:
                continue


T = typing.TypeVar("T")


def _chunk_generator(
    n: int, iterable: collections.abc.Iterable[T]
) -> collections.abc.Generator[collections.abc.Iterable[T], None, None]:
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)


# @typing.final
# class CartesianStrategy(interfaces.ExpansionStrategy):
#     """
#     Implements interfaces.ExpansionStrategy interface via Cartesian product of
#     molecules and operators.
#     """

#     def __init__(
#         self,
#         mol_lib: interfaces.ObjectLibrary[interfaces.MolDatBase],
#         op_lib: interfaces.ObjectLibrary[interfaces.OpDatBase],
#         rxn_lib: interfaces.ObjectLibrary[interfaces.RxnDatBase],
#         engine: _ReactionProvider,
#         blacklist: typing.Optional[set[interfaces.Identifier]] = None,
#     ) -> None:
#         """Initialize strategy by attaching libraries.

#         Parameters
#         ----------
#         mol_lib : interfaces.ObjectLibrary[interfaces.MolDatBase]
#             Library containing molecules.
#         op_lib : interfaces.ObjectLibrary[interfaces.OpDatBase]
#             Library containing operators.
#         rxn_lib : interfaces.ObjectLibrary[interfaces.RxnDatBase]
#             Library containing reactions.
#         engine : _ReactionProvider
#             Object used to initialize reactions.
#         """
#         self._mol_lib = mol_lib
#         self._op_lib = op_lib
#         self._rxn_lib = rxn_lib
#         self._engine = engine
#         if blacklist is None:
#             blacklist = set()
#         self._blacklist = blacklist

#         # stores active molecule IDs
#         self._active_mols: set[interfaces.Identifier] = set(
#             (uid for uid in self._mol_lib.ids())
#         )
#         # stores active operator IDs
#         self._active_ops: set[interfaces.Identifier] = set(
#             (uid for uid in self._op_lib.ids())
#         )
#         # stores combinations of reagents and operators which have been tried
#         self._recipe_cache: set[
#             tuple[interfaces.Identifier, tuple[interfaces.Identifier, ...]]
#         ] = set()

#         # create operator-molecule compatibility table
#         self._compat_table: dict[
#             interfaces.Identifier, list[list[interfaces.Identifier]]
#         ] = {}
#         for op_uid in self._active_ops:
#             self._add_op_to_compat(op_uid)

#         # fill operator-molecule compatibility table
#         for mol_uid in self._active_mols:
#             self._add_mol_to_compat(mol_uid)

#     def _add_mol_to_compat(
#         self, mol: typing.Union[interfaces.Identifier, interfaces.MolDatBase]
#     ) -> None:
#         """
#         Add entries to compat_table for new molecule.

#         Parameters
#         ----------
#         mol : interfaces.Identifier | interfaces.MolDatBase
#             Molecule to be added to compat_table.
#         """
#         if isinstance(mol, interfaces.MolDatBase):
#             mol_uid = mol.uid
#         else:
#             mol_uid = mol
#             mol = self._mol_lib[mol]

#         for op_uid in self._active_ops:
#             op = self._op_lib[op_uid]
#             for arg in range(len(self._compat_table[op_uid])):
#                 if op.compat(mol, arg):
#                     self._compat_table[op_uid][arg].append(mol_uid)

#     def _add_op_to_compat(
#         self, op: typing.Union[interfaces.Identifier, interfaces.OpDatBase]
#     ) -> None:
#         """
#         Add entries to compat_table for new operator.

#         Parameters
#         ----------
#         op : interfaces.Identifier | interfaces.OpDatBase
#             Operator to be added to compat_table.
#         """
#         if isinstance(op, interfaces.OpDatBase):
#             op_uid = op.uid
#         else:
#             op_uid = op
#             op = self._op_lib[op]

#         optable: list[list[interfaces.Identifier]] = [
#             [] for _ in range(len(op))
#         ]
#         for arg in range(len(optable)):
#             for mol_uid in self._active_mols:
#                 mol = self._mol_lib[mol_uid]
#                 if op.compat(mol, arg):
#                     optable[arg].append(mol_uid)
#         self._compat_table[op_uid] = optable

#     @property
#     def blacklist(self) -> set[interfaces.Identifier]:
#         return self._blacklist

#     @blacklist.setter
#     def blacklist(self, value: set[interfaces.Identifier]) -> None:
#         self._blacklist = value

#     def reset_recipe_cache(self) -> None:
#         self._recipe_cache = set()

#     def refresh(self) -> None:
#         if len(self._mol_lib) > len(self._active_mols):
#             for mol_uid in self._mol_lib.ids():
#                 if (
#                     mol_uid not in self._active_mols
#                     and mol_uid not in self._blacklist
#                 ):
#                     self._active_mols.add(mol_uid)
#                     self._add_mol_to_compat(mol_uid)

#         if len(self._op_lib) > len(self._active_ops):
#             for op_uid in self._op_lib.ids():
#                 if op_uid not in self._active_ops:
#                     self._active_ops.add(op_uid)
#                     self._add_op_to_compat(op_uid)

#     def expand(
#         self,
#         max_rxns: typing.Optional[int] = None,
#         max_mols: typing.Optional[int] = None,
#         num_gens: typing.Optional[int] = None,
#         custom_filter: typing.Optional[
#             collections.abc.Callable[
#                 [
#                     interfaces.OpDatBase,
#                     collections.abc.Sequence[interfaces.MolDatBase],
#                     collections.abc.Sequence[interfaces.MolDatBase],
#                 ],
#                 bool,
#             ]
#         ] = None,
#         custom_uid_prefilter: typing.Optional[
#             collections.abc.Callable[
#                 [
#                     interfaces.Identifier,
#                     collections.abc.Sequence[interfaces.Identifier],
#                 ],
#                 bool,
#             ]
#         ] = None,
#         retain_products_to_blacklist: bool = False,
#     ) -> None:
#         # value used to tell if any new reactions have occurred in a
#         # generation
#         exhausted: bool = False
#         num_mols: int = 0
#         num_rxns: int = 0
#         gen: int = 0

#         while not exhausted:
#             if num_gens is not None and gen >= num_gens:
#                 return
#             exhausted = True

#             # iterate through possible reactant combinations
#             for recipe in _generate_recipes_from_compat_table(
#                 self._compat_table, self._recipe_cache, custom_uid_prefilter
#             ):
#                 operator = self._op_lib[recipe[0]]
#                 reactants = tuple(
#                     (self._mol_lib[mol_uid] for mol_uid in recipe[1])
#                 )

#                 # iterate through evaluated reactions
#                 results, rejects = _evaluate_reaction(
#                     operator,
#                     reactants,
#                     self._engine,
#                     custom_filter,
#                     retain_products_to_blacklist,
#                 )
#                 for reaction, products in rejects:
#                     num_mols += len(products)
#                     num_rxns += 1
#                     if max_mols is not None and num_mols > max_mols:
#                         return
#                     if max_rxns is not None and num_rxns > max_rxns:
#                         return
#                     for mol in products:
#                         self._blacklist.add(mol.uid)
#                         self._mol_lib.add(mol)
#                     self._rxn_lib.add(reaction)
#                     exhausted = False
#                 for reaction, products in results:
#                     num_mols += len(products)
#                     num_rxns += 1
#                     if max_mols is not None and num_mols > max_mols:
#                         return
#                     if max_rxns is not None and num_rxns > max_rxns:
#                         return
#                     for mol in products:
#                         self._blacklist.discard(mol.uid)
#                         self._mol_lib.add(mol)
#                     self._rxn_lib.add(reaction)
#                     exhausted = False
#             gen += 1
#             self.refresh()


# @typing.final
# class CartesianStrategyParallel(interfaces.ExpansionStrategy):
#     """
#     Implements interfaces.ExpansionStrategy interface via Cartesian product of
#     molecules and operators.
#     """

#     def __init__(
#         self,
#         mol_lib: interfaces.ObjectLibrary[interfaces.MolDatBase],
#         op_lib: interfaces.ObjectLibrary[interfaces.OpDatBase],
#         rxn_lib: interfaces.ObjectLibrary[interfaces.RxnDatBase],
#         engine: _ReactionProvider,
#         num_procs: int,
#     ) -> None:
#         """Initialize strategy by attaching libraries.

#         Parameters
#         ----------
#         mol_lib : interfaces.ObjectLibrary[interfaces.MolDatBase]
#             Library containing molecules.
#         op_lib : interfaces.ObjectLibrary[interfaces.OpDatBase]
#             Library containing operators.
#         rxn_lib : interfaces.ObjectLibrary[interfaces.RxnDatBase]
#             Library containing reactions.
#         engine : _ReactionProvider
#             Object used to initialize reactions.
#         """
#         self._mol_lib = mol_lib
#         self._op_lib = op_lib
#         self._rxn_lib = rxn_lib
#         self._engine = engine
#         self._num_procs = num_procs

#         # stores active molecule IDs
#         self._active_mols: set[interfaces.Identifier] = set(
#             (uid for uid in self._mol_lib.ids())
#         )
#         # stores active operator IDs
#         self._active_ops: set[interfaces.Identifier] = set(
#             (uid for uid in self._op_lib.ids())
#         )
#         # stores combinations of reagents and operators which have been tried
#         self._recipe_cache: set[
#             tuple[interfaces.Identifier, tuple[interfaces.Identifier, ...]]
#         ] = set()

#         # create operator-molecule compatibility table
#         self._compat_table: dict[
#             interfaces.Identifier, list[list[interfaces.Identifier]]
#         ] = {}
#         for op_uid in self._active_ops:
#             self._add_op_to_compat(op_uid)

#         # fill operator-molecule compatibility table
#         for mol_uid in self._active_mols:
#             self._add_mol_to_compat(mol_uid)

#     def _add_mol_to_compat(
#         self, mol: typing.Union[interfaces.Identifier, interfaces.MolDatBase]
#     ) -> None:
#         """
#         Add entries to compat_table for new molecule.

#         Parameters
#         ----------
#         mol : interfaces.Identifier | interfaces.MolDatBase
#             Molecule to be added to compat_table.
#         """
#         if isinstance(mol, interfaces.MolDatBase):
#             mol_uid = mol.uid
#         else:
#             mol_uid = mol
#             mol = self._mol_lib[mol]

#         for op_uid in self._active_ops:
#             op = self._op_lib[op_uid]
#             for arg in range(len(self._compat_table[op_uid])):
#                 if op.compat(mol, arg):
#                     self._compat_table[op_uid][arg].append(mol_uid)

#     def _add_op_to_compat(
#         self, op: typing.Union[interfaces.Identifier, interfaces.OpDatBase]
#     ) -> None:
#         """
#         Add entries to compat_table for new operator.

#         Parameters
#         ----------
#         op : interfaces.Identifier | interfaces.OpDatBase
#             Operator to be added to compat_table.
#         """
#         if isinstance(op, interfaces.OpDatBase):
#             op_uid = op.uid
#         else:
#             op_uid = op
#             op = self._op_lib[op]

#         optable: list[list[interfaces.Identifier]] = [
#             [] for _ in range(len(op))
#         ]
#         for arg in range(len(optable)):
#             for mol_uid in self._active_mols:
#                 mol = self._mol_lib[mol_uid]
#                 if op.compat(mol, arg):
#                     optable[arg].append(mol_uid)
#         self._compat_table[op_uid] = optable

#     def reset_recipe_cache(self) -> None:
#         self._recipe_cache = set()

#     def refresh(self) -> None:
#         if len(self._mol_lib) > len(self._active_mols):
#             for mol_uid in self._mol_lib.ids():
#                 if mol_uid not in self._active_mols:
#                     self._active_mols.add(mol_uid)
#                     self._add_mol_to_compat(mol_uid)

#         if len(self._op_lib) > len(self._active_ops):
#             for op_uid in self._op_lib.ids():
#                 if op_uid not in self._active_ops:
#                     self._active_ops.add(op_uid)
#                     self._add_op_to_compat(op_uid)

#     def expand(
#         self,
#         max_rxns: typing.Optional[int] = None,
#         max_mols: typing.Optional[int] = None,
#         num_gens: typing.Optional[int] = None,
#         custom_filter: typing.Optional[
#             collections.abc.Callable[
#                 [
#                     interfaces.OpDatBase,
#                     collections.abc.Sequence[interfaces.MolDatBase],
#                     collections.abc.Sequence[interfaces.MolDatBase],
#                 ],
#                 bool,
#             ]
#         ] = None,
#         custom_uid_prefilter: typing.Optional[
#             collections.abc.Callable[
#                 [
#                     interfaces.Identifier,
#                     collections.abc.Sequence[interfaces.Identifier],
#                 ],
#                 bool,
#             ]
#         ] = None,
#     ) -> None:
#         # value used to tell if any new reactions have occurred in a
#         # generation
#         exhausted: bool = False
#         num_mols: int = 0
#         num_rxns: int = 0
#         gen: int = 0

#         while not exhausted:
#             if num_gens is not None and gen >= num_gens:
#                 return
#             exhausted = True

#             job_generator = _chunk_generator(
#                 self._num_procs * 100,
#                 (
#                     (
#                         self._op_lib[recipe[0]],
#                         [self._mol_lib[mol_uid] for mol_uid in recipe[1]],
#                         self._engine,
#                         custom_filter,
#                     )
#                     for recipe in _generate_recipes_from_compat_table(
#                         self._compat_table,
#                         self._recipe_cache,
#                         custom_uid_prefilter,
#                     )
#                 ),
#             )

#             # iterate through possible reactant combinations
#             with concurrent.futures.ProcessPoolExecutor(
#                 max_workers=self._num_procs
#             ) as executor:
#                 for jobchunk in job_generator:
#                     jobchunk = tuple(jobchunk)
#                     for results, _ in executor.map(
#                         _evaluate_reaction_unpack, jobchunk
#                     ):
#                         # for result in _evaluate_reaction_unpack(jobchunk):
#                         for reaction, products in results:
#                             num_mols += len(products)
#                             num_rxns += 1
#                             if max_mols is not None and num_mols > max_mols:
#                                 return
#                             if max_rxns is not None and num_rxns > max_rxns:
#                                 return
#                             for mol in products:
#                                 self._mol_lib.add(mol)
#                             self._rxn_lib.add(reaction)
#                             exhausted = False
#             gen += 1
#             self.refresh()


@dataclasses.dataclass(frozen=True)
class RecipeRankingJob:
    __slots__ = (
        "operator",
        "op_args",
        "mol_filter",
        "bundle_filter",
        "recipe_filter",
        "recipe_ranker",
        "heap_size",
    )

    operator: interfaces.DataPacket[interfaces.OpDatBase]
    op_args: tuple[
        tuple[interfaces.DataPacket[interfaces.MolDatBase], ...], ...
    ]
    mol_filter: typing.Optional[interfaces.MolFilter]
    bundle_filter: typing.Optional[interfaces.BundleFilter]
    recipe_filter: typing.Optional[interfaces.RecipeFilter]
    recipe_ranker: typing.Optional[interfaces.RecipeRanker]
    heap_size: typing.Optional[int]


def calc_batch_split(
    size_bundle: collections.abc.Sequence[int], batch_size: int
) -> tuple[int, ...]:
    num_split = list(itertools.repeat(1, len(size_bundle)))
    split_size = tuple(size_bundle)
    while math.prod(split_size) > batch_size:
        max_index = max(range(len(split_size)), key=split_size.__getitem__)
        num_split[max_index] += 1
        split_size = tuple(
            -(num_mols // -splitnum)
            for num_mols, splitnum in zip(size_bundle, num_split, strict=False)
        )
    return tuple(num_split)


def _generate_recipe_batches(
    mol_table: collections.abc.Sequence[
        collections.abc.Sequence[interfaces.MolIndex]
    ],
    table_indices: collections.abc.Sequence[int],
    batch_size: typing.Optional[int] = None,
    updated_mols: typing.Optional[set[interfaces.MolIndex]] = None,
) -> collections.abc.Generator[
    tuple[tuple[interfaces.MolIndex, ...], ...], None, None
]:
    if updated_mols is None:
        updated_mols = set()
    num_args = len(mol_table)

    # get number of old and new mols for each argument
    # num_old: list[int] = []
    # num_new: list[int] = []
    # for mols, i_counter in zip(mol_table, table_indices):
    #     n_new_from_old = len(updated_mols.intersection(mols[:i_counter]))
    #     n_new = len(updated_mols) - i_counter + n_new_from_old
    #     n_old = i_counter - n_new_from_old
    #     num_old.append(n_old)
    #     num_new.append(n_new)
    updated_vals: tuple[tuple[interfaces.MolIndex, ...], ...] = tuple(
        tuple(updated_mols.intersection(mol_list)) for mol_list in mol_table
    )
    num_old = tuple(table_indices)
    num_new = tuple(
        len(col) + len(mol_list) - i_counter
        for col, mol_list, i_counter in zip(
            updated_vals, mol_table, table_indices, strict=False
        )
    )
    num_tot = tuple(len(col) for col in mol_table)

    # for each argument, generate a bundle
    for i_bundle in range(num_args):
        # value_gens = (
        #     tuple(
        #         (i for i in mol_list[:i_counter] if i not in updated_mols)
        #         for mol_list, i_counter in zip(
        #             mol_table[:i_bundle], table_indices[:i_bundle]
        #         )
        #     )
        #     + ((i for i in mol_table[i_bundle]),)
        #     + tuple(
        #         (
        #             i
        #             for i in chain(
        #                 updated_mols.intersection(mol_list[:i_counter]),
        #                 mol_list[i_counter:],
        #             )
        #         )
        #         for mol_list, i_counter in zip(
        #             mol_table[i_bundle + 1 :], table_indices[i_bundle + 1 :]
        #         )
        #     )
        # )

        # # if no batch size, yield entire bundle as batch
        # if batch_size is None:
        #     yield tuple(tuple(mol_gen) for mol_gen in value_gens)
        #     return

        # # get bundle size
        # size_bundle = tuple(
        #     chain(
        #         num_old[:i_bundle],
        #         (num_new[i_bundle],),
        #         (
        #             n_new + n_old
        #             for n_new, n_old in zip(
        #                 num_new[i_bundle + 1 :], num_old[i_bundle + 1 :]
        #             )
        #         ),
        #     )
        # )
        if batch_size is None:
            yield (
                tuple(
                    tuple(mol_list[:i_counter])
                    for mol_list, i_counter in zip(
                        mol_table[:i_bundle], table_indices, strict=False
                    )
                )
                + (
                    updated_vals[i_bundle]
                    + tuple(mol_table[i_bundle][table_indices[i_bundle] :]),
                )
                + tuple(
                    tuple(mol_list) for mol_list in mol_table[i_bundle + 1 :]
                )
            )
            return

        size_bundle = (
            num_old[:i_bundle] + (num_new[i_bundle],) + num_tot[i_bundle + 1 :]
        )
        if not all(size_bundle):
            continue

        batch_split = calc_batch_split(size_bundle, batch_size)
        chunk_sizes = tuple(
            -(num_mols // -splitnum)
            for num_mols, splitnum in zip(
                size_bundle, batch_split, strict=False
            )
        )
        split_indices = (range(num_splits) for num_splits in batch_split)
        for batch_subindices in itertools.product(*split_indices):
            prev_column_mols = tuple(
                tuple(
                    mol_list[:i_counter][
                        subindex * chunk_size : (subindex + 1) * chunk_size
                    ]
                )
                for mol_list, i_counter, chunk_size, subindex in zip(
                    mol_table[:i_bundle],
                    table_indices[:i_bundle],
                    chunk_sizes[:i_bundle],
                    batch_subindices[:i_bundle],
                    strict=False,
                )
            )
            cur_column_mols = tuple(
                mol_table[i_bundle][
                    batch_subindices[i_bundle] * chunk_sizes[i_bundle] : (
                        batch_subindices[i_bundle] + 1
                    )
                    * chunk_sizes[i_bundle]
                ]
            )
            next_column_mols = tuple(
                tuple(
                    mol_list[
                        subindex * chunk_size : (subindex + 1) * chunk_size
                    ]
                )
                for mol_list, chunk_size, subindex in zip(
                    mol_table[i_bundle + 1 :],
                    chunk_sizes[i_bundle + 1 :],
                    batch_subindices[i_bundle + 1 :],
                    strict=False,
                )
            )
            yield prev_column_mols + (cur_column_mols,) + next_column_mols


def assemble_recipe_batch_job(
    op_index: interfaces.OpIndex,
    batch: tuple[tuple[interfaces.MolIndex, ...], ...],
    network: interfaces.ChemNetwork,
    keyset: interfaces.MetaKeyPacket,
    recipe_ranker: typing.Optional[interfaces.RecipeRanker] = None,
    heap_size: typing.Optional[int] = None,
    recipe_filter: typing.Optional[interfaces.RecipeFilter] = None,
    mol_filter: typing.Optional[interfaces.MolFilter] = None,
    bundle_filter: typing.Optional[interfaces.BundleFilter] = None,
) -> RecipeRankingJob:
    mol_data: typing.Union[
        collections.abc.Generator[
            collections.abc.Generator[None, None, None], None, None
        ],
        collections.abc.Generator[
            collections.abc.Generator[interfaces.MolDatBase, None, None],
            None,
            None,
        ],
    ]
    mol_meta: collections.abc.Generator[
        typing.Union[
            collections.abc.Sequence[collections.abc.Mapping],
            collections.abc.Generator[None, None, None],
        ],
        None,
        None,
    ]
    op_data = network.ops[op_index] if keyset.live_operator else None
    if keyset.operator_keys:
        op_meta = network.ops.meta(op_index, keyset.operator_keys)
    else:
        op_meta = None
    op = interfaces.DataPacket(op_index, op_data, op_meta)
    if keyset.live_molecule:
        mol_data = ((network.mols[i] for i in mol_list) for mol_list in batch)
    else:
        mol_data = ((None for _ in mol_list) for mol_list in batch)
    if keyset.molecule_keys:
        mol_meta = (
            tuple(network.mols.meta(mol_list, keyset.molecule_keys))
            for mol_list in batch
        )
    else:
        mol_meta = ((None for _ in mol_list) for mol_list in batch)
    mol_batch = tuple(
        tuple(
            interfaces.DataPacket(i, mol, m_meta)
            for i, mol, m_meta in zip(i_col, mol_col, meta_col, strict=False)
        )
        for i_col, mol_col, meta_col in zip(
            batch, mol_data, mol_meta, strict=False
        )
    )
    return RecipeRankingJob(
        op,
        mol_batch,
        mol_filter,
        bundle_filter,
        recipe_filter,
        recipe_ranker,
        heap_size,
    )


@dataclasses.dataclass(frozen=True)
class ReactionJob:
    __slots__ = (
        "operator",
        "op_args",
    )

    operator: interfaces.DataPacketE[interfaces.OpDatBase]
    op_args: tuple[interfaces.DataPacketE[interfaces.MolDatBase], ...]


def assemble_reaction_job(
    recipe: interfaces.Recipe,
    network: interfaces.ChemNetwork,
    keyset: interfaces.MetaKeyPacket,
) -> ReactionJob:
    op_index = recipe.operator
    op_data = network.ops[op_index]

    if keyset.operator_keys:
        op_meta = network.ops.meta(op_index, keyset.operator_keys)
    else:
        op_meta = None

    op = interfaces.DataPacketE(op_index, op_data, op_meta)

    mol_meta: collections.abc.Iterable[typing.Optional[collections.abc.Mapping]]
    if keyset.molecule_keys:
        mol_meta = network.mols.meta(recipe.reactants, keyset.molecule_keys)
    else:
        mol_meta = (None for _ in recipe.reactants)

    reactants = tuple(
        interfaces.DataPacketE(i, network.mols[i], meta)
        for i, meta in zip(recipe.reactants, mol_meta, strict=False)
    )

    return ReactionJob(op, reactants)


@dataclasses.dataclass(frozen=True)
class RecipePriorityItem:
    rank: typing.Optional[interfaces.RankValue]
    recipe: interfaces.Recipe

    def __lt__(self, other: "RecipePriorityItem") -> bool:
        if other.rank is None:
            if self.rank is None:
                return self.recipe < other.recipe
            return False
        if self.rank is None:
            return True
        return self.rank < other.rank

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, RecipePriorityItem)
            and self.rank == other.rank
            and self.recipe == other.recipe
        )


def execute_recipe_ranking(
    job: RecipeRankingJob,
    min_val: typing.Optional[RecipePriorityItem],
    recipes_tested: collections.abc.Collection[interfaces.Recipe],
) -> "RecipeHeap":
    # filter molecules
    args_edited: tuple[
        tuple[interfaces.DataPacket[interfaces.MolDatBase], ...], ...
    ]
    if job.mol_filter is not None:
        args_edited = tuple(
            tuple(
                mol for mol in arg_mols if job.mol_filter(mol, job.operator, i)
            )
            for i, arg_mols in enumerate(job.op_args)
        )
    else:
        args_edited = job.op_args

    # perform bundle filtering
    bundles: collections.abc.Iterable[interfaces.RecipeBundle]
    if job.bundle_filter is not None:
        bundles = job.bundle_filter(
            interfaces.RecipeBundle(job.operator, args_edited)
        )
    else:
        bundles = (interfaces.RecipeBundle(job.operator, args_edited),)

    recipe_generator = (
        (interfaces.RecipeExplicit(job.operator, reactants_data), recipe)
        for reactants_data, recipe in itertools.chain(
            *(
                (
                    (
                        reactants_data,
                        interfaces.Recipe(
                            interfaces.OpIndex(job.operator.i),
                            tuple(
                                interfaces.MolIndex(reactant.i)
                                for reactant in reactants_data
                            ),
                        ),
                    )
                    for reactants_data in itertools.product(*bundle.args)
                )
                for bundle in bundles
            )
        )
        if recipe not in recipes_tested
    )

    if job.recipe_ranker is None:
        if job.recipe_filter is None:
            return RecipeHeap.from_iter(
                (
                    recipe_item
                    for recipe_item in (
                        RecipePriorityItem(
                            None,
                            recipe,
                        )
                        for _, recipe in recipe_generator
                    )
                    if min_val is None or not (recipe_item < min_val)
                ),
                maxsize=job.heap_size,
            )
        return RecipeHeap.from_iter(
            (
                recipe_item
                for recipe_item in (
                    RecipePriorityItem(None, recipe)
                    for recipe_explicit, recipe in recipe_generator
                    if job.recipe_filter(recipe_explicit)
                )
                if min_val is None or not (recipe_item < min_val)
            ),
            job.heap_size,
        )

    # ideally, pass the min_rank to recipe_ranker, but only if the heap is full;
    # strategy utilizing heap and parallel reduction is likely best
    # but difficult to implement
    if job.recipe_filter is None:
        rank_generator = (
            RecipePriorityItem(job.recipe_ranker(recipe_explicit), recipe)
            for recipe_explicit, recipe in recipe_generator
        )
    else:
        rank_generator = (
            RecipePriorityItem(job.recipe_ranker(recipe_explicit), recipe)
            for recipe_explicit, recipe in recipe_generator
            if job.recipe_filter(recipe_explicit)
        )
    priority_item_generator = (
        recipe_item
        for recipe_item in rank_generator
        if (min_val is None) or not (recipe_item < min_val)
    )
    return RecipeHeap.from_iter(priority_item_generator, maxsize=job.heap_size)


class RecipeHeap:
    __slots__ = ("_heap", "_maxsize", "_ordered")

    _heap: typing.Optional[list[RecipePriorityItem]]
    _ordered: typing.Optional[list[RecipePriorityItem]]
    _maxsize: typing.Optional[int]

    def __init__(
        self,
        maxsize: typing.Optional[int] = None,
        heaps: typing.Optional[collections.abc.Collection["RecipeHeap"]] = None,
    ) -> None:
        if heaps is None:
            self._heap = None
            self._ordered = None
        elif maxsize is None:
            self._heap = None
            self._ordered = list(heapq.merge(*heaps))
        else:
            self._heap = None
            self._ordered = list(
                reversed(
                    tuple(
                        itertools.islice(
                            heapq.merge(
                                *(reversed(heap) for heap in heaps),
                                reverse=True,
                            ),
                            0,
                            maxsize,
                        )
                    )
                )
            )
        self._maxsize = maxsize

    @classmethod
    def from_iter(
        cls,
        data: collections.abc.Iterable[RecipePriorityItem],
        maxsize: typing.Optional[int] = None,
    ) -> "RecipeHeap":
        heap = RecipeHeap(maxsize)
        for item in data:
            heap.add_recipe(item)
        return heap

    @property
    def min(self) -> typing.Optional[RecipePriorityItem]:
        if self._heap is None:
            if self._ordered is None or len(self._ordered) == 0:
                return None
            return min(self._ordered)
        return self._heap[0]

    def popvals(
        self, n: typing.Optional[int]
    ) -> tuple[RecipePriorityItem, ...]:
        if self._ordered is None:
            if self._heap is None:
                return tuple()
            self._ordered = sorted(self._heap)
        self._heap = None
        if n is None:
            vals = tuple(reversed(self._ordered))
            self._ordered = None
            return vals
        return tuple(
            self._ordered.pop() for _ in range(min(len(self._ordered), n))
        )

    @typing.overload
    def __getitem__(self, item: slice) -> list[RecipePriorityItem]: ...

    @typing.overload
    def __getitem__(self, item: int) -> RecipePriorityItem: ...

    def __getitem__(self, item: typing.Union[int, slice]):
        if self._ordered is None:
            if self._heap is None:
                raise IndexError(f"Invalid index {item=}; heap is empty")
            self._ordered = sorted(self._heap)
        return self._ordered[item]

    def __len__(self) -> int:
        if self._heap is None:
            if self._ordered is None:
                return 0
            return len(self._ordered)
        return len(self._heap)

    def __iter__(self) -> collections.abc.Iterator[RecipePriorityItem]:
        if self._ordered is None:
            if self._heap is None:
                return iter([])
            self._ordered = sorted(self._heap)
        return iter(self._ordered)

    def __reversed__(self) -> collections.abc.Iterator[RecipePriorityItem]:
        if self._ordered is None:
            if self._heap is None:
                return iter([])
            self._ordered = sorted(self._heap)
        return reversed(self._ordered)

    def add_recipe(self, recipe: RecipePriorityItem) -> None:
        if self._heap is None:
            if self._ordered is None:
                self._heap = []
            else:
                self._heap = list(self._ordered)
                heapq.heapify(self._heap)
                self._ordered = None
        if self._maxsize is None or len(self._heap) < self._maxsize:
            heapq.heappush(self._heap, recipe)
        elif self._heap[0] < recipe:
            heapq.heapreplace(self._heap, recipe)

    def __add__(self, other: "RecipeHeap") -> "RecipeHeap":
        if self._maxsize != other._maxsize:
            raise ValueError(
                f"Heap sizes do not match ({self._maxsize} != {other._maxsize})"
            )

        return RecipeHeap(self._maxsize, (self, other))


def execute_reaction(
    rxn_job: ReactionJob,
) -> collections.abc.Generator[
    tuple[interfaces.ReactionExplicit, bool], None, None
]:
    reactants = tuple(mol.item for mol in rxn_job.op_args)
    if rxn_job.operator.item is None or any(mol is None for mol in reactants):
        raise ValueError("ReactionJob has non-None item components!")
    product_packets = rxn_job.operator.item(*reactants)
    reactant_datapackets = tuple(
        interfaces.DataPacketE(p.i, p.item, p.meta) for p in rxn_job.op_args
    )
    operator_datapacket = interfaces.DataPacketE(
        rxn_job.operator.i, rxn_job.operator.item, rxn_job.operator.meta
    )
    for rxn_products in product_packets:
        product_datapackets = tuple(
            interfaces.DataPacketE(-1, product, None)
            for product in rxn_products
        )
        rxn = interfaces.ReactionExplicit(
            operator_datapacket,
            reactant_datapackets,
            product_datapackets,
            None,
        )
        yield rxn, True


def execute_reactions(
    rxn_jobs: collections.abc.Collection[ReactionJob],
    rxn_analysis: typing.Optional[metadata.RxnAnalysisStep] = None,
) -> collections.abc.Iterable[tuple[interfaces.ReactionExplicit, bool]]:
    rxn_generator = itertools.chain.from_iterable(
        execute_reaction(rxn_job) for rxn_job in rxn_jobs
    )
    if rxn_analysis is None:
        return rxn_generator
    return rxn_analysis.execute(rxn_generator)


class PriorityQueueStrategyBasic(interfaces.PriorityQueueStrategy):
    __slots__ = ("_network",)

    def __init__(
        self,
        network: interfaces.ChemNetwork,
        num_procs: typing.Optional[int] = None,
    ) -> None:
        if num_procs is not None:
            raise NotImplementedError(
                f"Parallel processing not yet supported on {type(self)}"
            )
        self._network = network

    def expand(
        self,
        max_recipes: typing.Optional[int] = None,
        mol_filter: typing.Optional[interfaces.MolFilter] = None,
        bundle_filter: typing.Optional[interfaces.BundleFilter] = None,
        recipe_filter: typing.Optional[interfaces.RecipeFilter] = None,
        recipe_ranker: typing.Optional[interfaces.RecipeRanker] = None,
        reaction_plan: typing.Optional[
            typing.Union[
                metadata.RxnAnalysisStep,
                metadata.PropertyCompositor,
                metadata.ReactionFilterBase,
                metadata.LocalPropertyCalc,
            ]
        ] = None,
        global_hooks: typing.Optional[
            collections.abc.Sequence[interfaces.GlobalUpdateHook]
        ] = None,
        heap_size: typing.Optional[int] = None,
        beam_size: typing.Optional[int] = 1,
        batch_size: typing.Optional[int] = None,
        save_unreactive: bool = True,
    ) -> None:
        rxn_analysis_task: typing.Optional[metadata.RxnAnalysisStep] = None
        if reaction_plan is not None:
            rxn_analysis_task = metadata.as_rxn_analysis_step(reaction_plan)
            mc_update = rxn_analysis_task.resolver
        else:
            mc_update = metadata.MetaUpdateResolver({}, {}, {})

        if heap_size is not None and beam_size is not None:
            raise ValueError(
                f"""Heap size ({heap_size}) must be greater than beam size
                    ({beam_size})"""
            )

        # set keysets so that updated reactions may occur and parameters may be
        # passed to parallel processes
        recipe_keyset: interfaces.MetaKeyPacket = interfaces.MetaKeyPacket()
        reaction_keyset: interfaces.MetaKeyPacket = interfaces.MetaKeyPacket()
        if mol_filter is not None:
            recipe_keyset = recipe_keyset + mol_filter.meta_required
        if bundle_filter is not None:
            recipe_keyset = recipe_keyset + bundle_filter.meta_required
        if recipe_filter is not None:
            recipe_keyset = recipe_keyset + recipe_filter.meta_required
        if recipe_ranker is not None:
            recipe_keyset = recipe_keyset + recipe_ranker.meta_required
        if rxn_analysis_task is not None:
            reaction_keyset = reaction_keyset + rxn_analysis_task.meta_required
        total_keyset = recipe_keyset + reaction_keyset

        # initialize loop variables
        network = self._network
        compat_indices_table = [
            [0 for _ in network.compat_table(interfaces.OpIndex(i))]
            for i in range(len(network.ops))
        ]
        updated_mols_set: set[interfaces.MolIndex] = set()
        updated_ops_set: set[interfaces.OpIndex] = set()
        recipe_heap: RecipeHeap = RecipeHeap(maxsize=heap_size)
        recipes_tested: set[interfaces.Recipe] = set()

        while (max_recipes is None or max_recipes > 0) and (
            any(
                any(
                    i_tested < len(compat_mols)
                    for i_tested, compat_mols in zip(
                        op_index_table,
                        network.compat_table(interfaces.OpIndex(op_index)),
                        strict=False,
                    )
                )
                for op_index, op_index_table in enumerate(compat_indices_table)
            )
            or len(updated_mols_set) != 0
            or len(updated_ops_set) != 0
            or len(recipe_heap) > 0
        ):
            for op_index in updated_ops_set:
                compat_indices_table[op_index] = [
                    0 for _ in range(len(compat_indices_table[op_index]))
                ]

            # for each operator, create recipe batches
            cur_min = recipe_heap.min
            for opIndex, _ in enumerate(network.ops):
                # for each argument, accumulate a total of old_mols and new_mols
                compat_table = network.compat_table(interfaces.OpIndex(opIndex))
                compat_indices = compat_indices_table[opIndex]
                if not all(compat_table):
                    continue
                # generate recipe batches
                for batch in _generate_recipe_batches(
                    compat_table, compat_indices, batch_size, updated_mols_set
                ):
                    # assemble recipe ranking job
                    recipejob = assemble_recipe_batch_job(
                        interfaces.OpIndex(opIndex),
                        batch,
                        network,
                        recipe_keyset,
                        recipe_ranker,
                        heap_size,
                        recipe_filter,
                        mol_filter,
                        bundle_filter,
                    )
                    new_recipe_heap = execute_recipe_ranking(
                        recipejob, cur_min, recipes_tested
                    )
                    recipe_heap = recipe_heap + new_recipe_heap

                continue

            # update compat_indices_table
            compat_indices_table = [
                [
                    len(mol_list)
                    for mol_list in network.compat_table(interfaces.OpIndex(i))
                ]
                for i in range(len(network.ops))
            ]

            # perform beam expansion
            recipes_to_be_expanded = recipe_heap.popvals(beam_size)

            if len(recipes_to_be_expanded) == 0:
                break

            if max_recipes is not None:
                if len(recipes_to_be_expanded) > max_recipes:
                    recipes_to_be_expanded = recipes_to_be_expanded[
                        :max_recipes
                    ]
                max_recipes = max_recipes - len(recipes_to_be_expanded)

            reaction_jobs = tuple(
                assemble_reaction_job(
                    reciperank.recipe, network, reaction_keyset
                )
                for reciperank in recipes_to_be_expanded
            )

            # execute reactions
            for rxn, pass_filter in execute_reactions(
                reaction_jobs, rxn_analysis_task
            ):
                if not save_unreactive and not pass_filter:
                    continue
                # add product mols to network
                products_indices = tuple(
                    network.add_mol(mol.item, None, pass_filter)
                    for mol in rxn.products
                    if mol.item is not None
                )

                # build reaction
                reactants_indices = tuple(
                    interfaces.MolIndex(mol.i) for mol in rxn.reactants
                )
                rxn_implicit = interfaces.Reaction(
                    interfaces.OpIndex(rxn.operator.i),
                    reactants_indices,
                    products_indices,
                )

                # add reaction to network
                rxn_index = network.add_rxn(rxn=rxn_implicit)

                updated_mols_set = set()
                updated_ops_set = set()

                # update reactant metadata
                key: collections.abc.Hashable
                for m_dat in zip(
                    reactants_indices, rxn.reactants, strict=False
                ):
                    if m_dat[1].meta is not None:
                        cur_vals = network.mols.meta(m_dat[0], m_dat[1].meta)
                        for key, v in m_dat[1].meta.items():
                            value = v
                            if key in mc_update.mol_updates and key in cur_vals:
                                value = mc_update.mol_updates[key](
                                    value, cur_vals[key]
                                )
                            if key not in cur_vals or cur_vals[key] != value:
                                network.mols.set_meta(m_dat[0], {key: value})
                                if key in total_keyset.molecule_keys:
                                    updated_mols_set.add(m_dat[0])

                # update product metadata
                for m_dat in zip(products_indices, rxn.products, strict=False):
                    if m_dat[1].meta is not None:
                        cur_vals = network.mols.meta(m_dat[0], m_dat[1].meta)
                        for key, v in m_dat[1].meta.items():
                            value = v
                            if key in mc_update.mol_updates and key in cur_vals:
                                value = mc_update.mol_updates[key](
                                    value, cur_vals[key]
                                )
                            if key not in cur_vals or cur_vals[key] != value:
                                network.mols.set_meta(m_dat[0], {key: value})
                                if key in total_keyset.molecule_keys:
                                    updated_mols_set.add(m_dat[0])

                # update operator metadata
                if rxn.operator.meta is not None:
                    cur_vals = network.ops.meta(
                        rxn_implicit.operator, rxn.operator.meta
                    )
                    for key, v in rxn.operator.meta.items():
                        value = v
                        if key in mc_update.op_updates and key in cur_vals:
                            value = mc_update.op_updates[key](
                                value,
                                cur_vals[key],
                            )
                        if key not in cur_vals or cur_vals[key] != value:
                            network.ops.set_meta(
                                rxn_implicit.operator, {key: value}
                            )
                            if key in total_keyset.operator_keys:
                                updated_ops_set.add(rxn_implicit.operator)

                # update reaction metadata
                if rxn.reaction_meta is not None:
                    cur_vals = network.rxns.meta(rxn_index, rxn.reaction_meta)
                    for key, v in rxn.reaction_meta.items():
                        value = v
                        if key in mc_update.rxn_updates and key in cur_vals:
                            value = mc_update.rxn_updates[key](
                                value, cur_vals[key]
                            )
                        if key not in cur_vals or cur_vals[key] != value:
                            network.rxns.set_meta(rxn_index, {key: value})

            recipes_tested.update(
                (reciperank.recipe for reciperank in recipes_to_be_expanded)
            )

            if len(recipe_heap) == 0:
                compat_indices_table = [
                    [0 for _ in network.compat_table(interfaces.OpIndex(i))]
                    for i in range(len(network.ops))
                ]

            # run global hooks
            if global_hooks is not None:
                fake_network = pgnetworks.ChemNetworkFacadeMetaTrigger(
                    self._network, total_keyset
                )
                end_on_completion: bool = False
                for hook_func in global_hooks:
                    rval = hook_func(fake_network)
                    if (
                        rval
                        is interfaces.GlobalHookReturnValue.STOP_SHORTCIRCUIT
                    ):
                        return
                    if rval is interfaces.GlobalHookReturnValue.STOP:
                        end_on_completion = True
                if end_on_completion:
                    return

            continue


class CartesianStrategyUpdated:
    def __init__(
        self,
        network: interfaces.ChemNetwork,
        engine: interfaces.NetworkEngine,
        #        gen_key: collections.abc.Hashable = "generation",
    ) -> None:
        self._network = network
        self._engine = engine

    #        self._gen_key = gen_key

    def expand(
        self,
        num_iter: typing.Optional[int] = None,
        max_recipes: typing.Optional[int] = None,
        mol_filter: typing.Optional[interfaces.MolFilter] = None,
        bundle_filter: typing.Optional[interfaces.BundleFilter] = None,
        recipe_filter: typing.Optional[interfaces.RecipeFilter] = None,
        reaction_plan: typing.Optional[
            typing.Union[
                metadata.RxnAnalysisStep,
                metadata.PropertyCompositor,
                metadata.ReactionFilterBase,
                metadata.LocalPropertyCalc,
            ]
        ] = None,
        global_hooks: typing.Optional[
            collections.abc.Sequence[interfaces.GlobalUpdateHook]
        ] = None,
        save_unreactive: bool = True,
        # max_gen: typing.Optional[int] = None,
    ):
        engine = self._engine
        p_strat = engine.strat.pq(self._network)
        # mol_filter: typing.Optional[interfaces.MolFilter] = None
        # calc: typing.Optional[metadata.MolPropertyFromRxnCalc] = None
        # resolver: typing.Optional[metadata.MetaUpdateResolver] = None
        global_hook: list[interfaces.GlobalUpdateHook]
        global_hook = [] if global_hooks is None else list(global_hooks)
        # if max_gen is not None:
        #     mol_filter = engine.filter.mol.meta_func(
        #         self._gen_key, self.gen_test(max_gen)
        #     )
        #     calc = engine.meta.generation(self._gen_key)
        #     resolver = metadata.MetaUpdateResolver({self._gen_key: min},
        #                                            {}, {})
        if num_iter is not None:
            global_hook.append(engine.hook.max_iter(num_iter))
        p_strat.expand(
            max_recipes,
            # mol_filter,
            # reaction_plan=calc,
            # mc_update=resolver,
            beam_size=None,
            global_hooks=global_hook,
            mol_filter=mol_filter,
            bundle_filter=bundle_filter,
            recipe_filter=recipe_filter,
            reaction_plan=reaction_plan,
            save_unreactive=save_unreactive,
        )



"""Sphinx documentation builder."""

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import pathlib
import sys

sys.path.insert(0, pathlib.Path(__file__).parents[2].resolve().as_posix())

project = "DORAnet"
copyright = "2024, William Sprague"
author = "William Sprague"
release = "0.4.1a1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "numpydoc",
    "sphinxcontrib.mermaid",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autosummary",
]
autosummary_generate = True
myst_fence_as_directive = ["mermaid"]
numpydoc_show_class_members = False

templates_path = ["_templates"]
exclude_patterns: list[str] = []

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
